import re
import codecs
from os.path import isfile, exists
from .lua_globals import *
from . import lua_table, lua_math

version = [1, 0, 0]

# reserved names
_reserved = [
	"return",
	"break",
	"function",
	"local",
	"for",
	"repeat",
	"while",
	"until",
	"do",
	"and",
	"or",
	"not",
	"end",
	"true",
	"false",
	"if",
	"elseif",
	"else",
	"in",
	"nil",
]

def _getTableVar(var, args):
	if len(args) > 0:
		if not isinstance(var, dict):
			raise ValueError("attempt to index a %s value" % (luaType(var)))
		
		newVar = None
		if args[0] in var:
			newVar = var[args[0]]
		
		if len(args) == 1:
			return newVar
		elif len(args) > 1:
			return _getTableVar(newVar, args[1:])

class luaVarScope():
	def __init__(self, vars={}, parent=None):
		self.__vars = vars.copy()
		self.__parent = parent
		self.__children = []
		self.__lastVar = {}
	
	def __del__(self):
		for child in self.__children:
			del child
		
	def __setTableVar(self, var, args, value):
		if len(args) > 0:
			if not isinstance(var, dict):
				raise ValueError("attempt to index a %s value (global '%s')" % (luaType(var), baseName))
			
			if len(args) == 1:
				var[args[0]] = value
			elif len(args) > 1:
				self.__setTableVar(var[args[0]], args[1:], value)
	
	def getLastVar(self):
		return self.__lastVar
	
	def setLastVar(self, lastVar):
		self.__lastVar = lastVar

	def getChild(self, vars={}):
		child = luaVarScope(vars, self)
		self.__children.append(child)
		return child
		
	def reset(self, vars={}):
		self.__vars = vars.copy()
	
	def getVars(self):
		return self.__vars
	
	def setVar(self, varName, value, isLocal = False, args=[]):
		argList = args.copy()
		for i in range(len(argList)):
			if isinstance(argList[i], luaValue):
				argList[i] = argList[i].getValue(self)
		
		if isLocal:
			self.__vars[varName] = value
		else:
			if varName in self.__vars:
				# local var found
				if len(argList) > 0:
					self.__setTableVar(self.__vars[varName], argList, value)
				else:
					self.__vars[varName] = value
			
			elif isinstance(self.__parent, luaVarScope):
				# search var in parent
				self.__parent.setVar(varName, value, False, argList)
				
			else:
				# we're already in global scope
				if len(args) > 0:
					# can't index non initialized variables
					raise ValueError("attempt to index a nil value (global '%s')" % varName)
				else:
					self.__vars[varName] = value

	def getVar(self, varName, args=[]):
		argList = args.copy()
		for i in range(len(argList)):
			if isinstance(argList[i], luaValue):
				argList[i] = argList[i].getValue(self)
		
		var = None
		self.__lastVar = {"scope": "global", "varName": varName}
		if varName in self.__vars:
			# local var found
			if len(argList) > 0:
				var = _getTableVar(self.__vars[varName], argList)
			else:
				var = self.__vars[varName]
			
			if isinstance(self.__parent, luaVarScope):
				self.__lastVar["scope"] = "local"
		
		elif isinstance(self.__parent, luaVarScope):
			var = self.__parent.getVar(varName, argList)
			self.__lastVar = self.__parent.getLastVar()
		
		return var
	
	def showVars(self):
		print(self.__vars)

class luaBreak():
	# empty class to indicate break statement
	pass

class luaReturn():
	def __init__(self, value):
		self.__value = value
	
	def getValue(self):
		return self.__value
	
class luaValue(object):
	# base class for all values with mutable content
	def getValue(self):
		return None

class luaFormula(luaValue):
	# to-do: check if operation is allowed	
	def __init__(self, operator, value1, value2=None):
		self.__operators = {
			"(": (self.__parenthesis, 0),
			"^": (self.__power, 1),
			"not": (self.__not, 2),
			"#": (self.__length, 2),
			"*": (self.__mul, 3),
			"/": (self.__div, 3),
			"//": (self.__div2, 3),
			"%": (self.__mod, 3),
			"+": (self.__add, 4),
			"-": (self.__sub, 4, 2),
			"..": (self.__stringadd, 5),
			"<": (self.__lesser, 6),
			">": (self.__greater, 6),
			"<=": (self.__lesser_equal, 6),
			">=": (self.__greater_equal, 6),
			"~=": (self.__unequal, 6),
			"==": (self.__equal, 6),
			"and": (self.__and, 7),
			"or": (self.__or, 7),
		}
		
		self.__operator = operator	#string
		self.__value1 = value1		#luaVaribale or any python var type
		self.__value2 = value2		#luaVaribale or any python var type
		
		if isinstance(value2, luaFormula):
			if self.__getPriority() <= value2.__getPriority():
				# switch formulas
				self.__operator = value2.__operator
				self.__value1 = luaFormula(operator, value1, value2.__value1)
				self.__value2 = value2.__value2
		
		if isinstance(value1, luaFormula):
			if self.__getPriority() > 0 and self.__getPriority() <= value1.__getPriority():
				# switch formulas
				self.__operator = value1.__operator
				self.__value1 = luaFormula(operator, value1.__value1, value2)
				self.__value2 = value1.__value2
	
	def __getValue(self, valueStr, varScope):
		if valueStr == "__value2":
			value = self.__value2
		else:
			value = self.__value1
		
		if isinstance(value, luaValue):
			return value.getValue(varScope), varScope.getLastVar()
		
		return value, {}

	def __compareable(self, value1, value2):
		value1Type = luaType(value1)
		value2Type = luaType(value2)
		if value1Type != value2Type:
			raise TypeError("attempt to compare %s with %s" % (value1Type, value2Type))
		elif value1Type not in ("number", "string"):
			raise TypeError("attempt to compare two %s values" % value1Type)
		else:
			return True
	
	def __isNumber(self, value1, value2):
		valueType = luaType(value1[0])
		debugInfo = value1[1]
		if valueType != "number":
			errorMessage = "attempt to perform arithmetic on a %s value" % valueType
			if "scope" in debugInfo and "varName" in debugInfo:
				errorMessage += "(%s '%s')" % (debugInfo["scope"], debugInfo["varName"])
			raise TypeError(errorMessage)
		
		valueType = luaType(value2[0])
		debugInfo = value2[1]
		if valueType != "number":
			errorMessage = "attempt to perform arithmetic on a %s value" % valueType
			if "scope" in debugInfo and "varName" in debugInfo:
				errorMessage += "(%s '%s')" % (debugInfo["scope"], debugInfo["varName"])
			raise TypeError(errorMessage)
		
		return True

	def __parenthesis(self, varScope):
		value, debugInfo = self.__getValue("__value1", varScope)
		return value
		
	def __add(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 + value2
		
	def __stringadd(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		valueType = luaType(value1)
		if valueType not in ("number", "string"):
			errorMessage = "attempt to concatenate a %s value" % valueType
			if "scope" in debugInfo and "varName" in debugInfo:
				errorMessage += "(%s '%s')" % (debugInfo["scope"], debugInfo["varName"])
			raise TypeError(errorMessage)
		
		value2, debugInfo = self.__getValue("__value2", varScope)
		valueType = luaType(value2)
		if valueType not in ("number", "string"):
			errorMessage = "attempt to concatenate a %s value" % valueType
			if "scope" in debugInfo and "varName" in debugInfo:
				errorMessage += "(%s '%s')" % (debugInfo["scope"], debugInfo["varName"])
			raise TypeError(errorMessage)

		return toString(value1) + toString(value2)
		
	def __sub(self, varScope):
		if self.__value2 != None:
			value1, debugInfo1 = self.__getValue("__value1", varScope)
			value2, debugInfo2 = self.__getValue("__value2", varScope)
			if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
				return value1 - value2
		else:
			value1, debugInfo1 = self.__getValue("__value1", varScope)
			value2, debugInfo2 = 0, {}
			if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
				return -value1
		
	def __mul(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 * value2
		
	def __div(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 / value2
		
	def __div2(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 // value2
		
	def __and(self, varScope):
		value, debugInfo = self.__getValue("__value1", varScope)
		if not toBool(value):
			return value
		else:
			value, debugInfo = self.__getValue("__value2", varScope)
			return value
		
	def __or(self, varScope):
		value, debugInfo = self.__getValue("__value1", varScope)
		if toBool(value):
			return value
		else:
			value, debugInfo = self.__getValue("__value2", varScope)
			return value
		
	def __not(self, varScope):
		value, debugInfo = self.__getValue("__value1", varScope)
		return not toBool(value)
		
	def __length(self, varScope):
		value, debugInfo = self.__getValue("__value1", varScope)
		if isinstance(value, list):
			return len(value)+1
		if isinstance(value, str):
			return len(value)
		elif isinstance(value, dict):
			return lua_table.table["maxn"](value)
		else:
			errorMessage = "attempt to get length of a %s value" % luaType(value)
			if "scope" in debugInfo and "varName" in debugInfo:
				errorMessage += " (%s '%s')" % (debugInfo["scope"], debugInfo["varName"])
			raise TypeError(errorMessage)
		
	def __mod(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 % value2
		
	def __lesser(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		if self.__compareable(value1, value2):
			return value1 < value2
		
	def __greater(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		if self.__compareable(value1, value2):
			return value1 > value2
		
	def __lesser_equal(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		if self.__compareable(value1, value2):
			return value1 <= value2
		
	def __greater_equal(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		if self.__compareable(value1, value2):
			return value1 >= value2
		
	def __unequal(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		return value1 != value2
		
	def __equal(self, varScope):
		value1, debugInfo = self.__getValue("__value1", varScope)
		value2, debugInfo = self.__getValue("__value2", varScope)
		return value1 == value2
		
	def __power(self, varScope):
		value1, debugInfo1 = self.__getValue("__value1", varScope)
		value2, debugInfo2 = self.__getValue("__value2", varScope)
		if self.__isNumber((value1, debugInfo1), (value2, debugInfo2)):
			return value1 ** value2
	
	def __getPriority(self):
		if self.__operator in self.__operators:
			if self.__value2 == None and len(self.__operators[self.__operator]) > 2:
				return self.__operators[self.__operator][2]
			else:
				return self.__operators[self.__operator][1]
		return -1
	
	def getValue(self, varScope):
		if self.__operator in self.__operators:
			return self.__operators[self.__operator][0](varScope)
		
		return None
	
class luaVariable(luaValue):
	def __init__(self, name, args):
		self.__name = name	#string
		self.__args = args	#list
		
	def getValue(self, varScope):
		return varScope.getVar(self.__name, self.__args)
	
class luaFunctionCall(luaValue):
	def __init__(self, function, args, args2=[]):
		self.__function = function	#luaVariable
		self.__args = args 			#list
		self.__args2 = args2			#list
		
	def getValue(self, varScope):
		function = self.__function.getValue(varScope)
		if luaType(function) != "function":
			# not callable
			debug = varScope.getLastVar()
			raise TypeError("attempt to call a %s value (%s '%s')" % (luaType(function), debug["scope"], debug["varName"]))
		
		args = []
		argList = self.__args.copy()
		i = 0
		while i < len(argList):
			arg = argList[i]
			if isinstance(arg, luaValue):
				arg = arg.getValue(varScope)
				if isinstance(arg, list):
					if i == len(argList) - 1:
						for j in range(1, len(arg)):
							argList.append(arg[j])
					arg = arg[0]
				args.append(arg)
			else:
				args.append(arg)
			i += 1
		
		if len(args) > 0:
			result = function(luaArgList(args))
		else:
			result = function()
		
		args2 = self.__args2.copy()
		for i in range(len(args2)):
			if isinstance(args2[i], luaValue):
				args2[i] = args2[i].getValue(varScope)
		
		if len(args2) > 0:
			if isinstance(result, list):
				if isinstance(args2[0], int):
					result = result[args2[0]]
					args2 = args2[1:]
				else:
					result = result[0]
			
			result = _getTableVar(result, args2)
		
		return result
	
class luaTable(luaValue):
	def __init__(self):
		self.__keyValues = {}
		self.__values = []
		self.__maxIndex = 0

	def add_value(self, value):			
		self.__values.append(value)

	def add_key_value(self, value, key):			
		self.__keyValues[key] = value
	
	def getValue(self, varScope):
		table = {}
		
		# key, value pairs
		for key in self.__keyValues:
			value = self.__keyValues[key]
			if isinstance(value, luaValue):
				value = value.getValue(varScope)
			if isinstance(value, list):
				value = value[0]
			if isinstance(key, luaValue):
				key = key.getValue(varScope)
				if isinstance(key, list):
					key = key[0]
				elif key == None:
					raise ValueError("table index is nil")
			
			table[key] = value
		
		# list values, may override numerical keys!
		i = 0
		values = self.__values.copy()
		while i < len(values):
			key = i + 1
			value = values[i]
			if isinstance(value, luaValue):
				value = value.getValue(varScope)
			if isinstance(value, list):
				if key == len(values):
					for j in range(1, len(value)):
						values.append(value[j])
					value = value[0]

			table[key] = value
			i += 1

		return table

class luaIterator(luaValue):
	def __init__(self, start, end, step):
		self.__start = start
		self.__end = end
		self.__step = step
		
	def getValue(self, varScope):
		start = self.__start
		if isinstance(start, luaValue):
			start = start.getValue(varScope)
		end = self.__end
		if isinstance(end, luaValue):
			end = end.getValue(varScope)
		step = self.__step
		if isinstance(step, luaValue):
			step = step.getValue(varScope)
		
		return range(start, end+1, step)
		
class luaExpression():
	def __init__(self, type, value=None, varName="", isLocal=False, args=[]):
		self.type = type
		self.__value = value
		self.__varName = varName
		self.__isLocal = isLocal
		self.__args = args

	def execute(self, varScope):
		if self.type == "return":
			result = self.__value.copy()
			i = 0
			while i < len(result):
				if isinstance(result[i], luaValue):
					result[i] = result[i].getValue(varScope)
				if isinstance(result[i], list):
					if i == len(result)-1:
						for j in range(1, len(result[i])):
							result.append(result[i][j])
					result[i] = result[i][0]
				i += 1					
			return result
		elif self.type == "assign_var":
			values = self.__value.copy()
			for i in range(len(self.__varName)):
				varName = self.__varName[i]
				args = self.__args[i]
				value = None
				if i < len(values):
					value = values[i]
					if isinstance(value, luaFunctionCall):
						value = value.getValue(varScope)
					if isinstance(value, list):
						if i == len(values)-1:
							for j in range(1, len(value)):
								values.append(value[j])
						value = value[0]
				
				if isinstance(value, luaValue):
					value = value.getValue(varScope)
				varScope.setVar(varName, value, self.__isLocal, args)
		elif self.type == "function_call":
			result = self.__value
			if isinstance(result, luaValue):
				result = result.getValue(varScope)
		elif self.type == "control_structure":
			result = self.__value.execute(varScope)
			return result
		
		return None

class luaControlStructure():
	def __init__(self, text, parent=None, filename=""):
		self._luaText = text
		self.__filename = filename
		if filename != "":
			self.__fullText = text
		else:
			self.__fullText = ""
		
		self._parent = parent
		if isinstance(self._parent, luaControlStructure):
			self._insideLoop = self._parent._insideLoop
		else:
			self._insideLoop = False
		
		self._expressions = []
		
		self._patterns = {
			"string": r'("|\[([=]*)\[|\')',
			"varName": r'[a-zA-Z0-9_]+',
			"varName2": r'[a-zA-Z0-9_]+|[.]{3}',
			"operator": r'\^|\+|-|\*|//|/|and|or|not|#|%|\.\.|<=|>=|<|>|~=|==',
		}
		
		self.__base_expressions = [
			{"pattern": r'\s*--', "function": self.__read_comment},
			{"pattern": r'\s*return\W', "function": self.__read_return},
			{"pattern": r'\s*break\W', "function": self.__read_break},
			{"pattern": r'\s*for\W', "function": self.__read_for_loop},
			{"pattern": r'\s*if\W', "function": self.__read_if_statement},
			{"pattern": r'\s*while\W', "function": self.__read_while_loop},
			{"pattern": r'\s*repeat\W', "function": self.__read_repeat_loop},
			{"pattern": r'\s*do\W', "function": self.__read_do_block},
			{"pattern": r'\s*local\W', "function": self.__read_local},
			{"pattern": r'\s*function\W', "function": self.__read_named_function},
			{"pattern": r'\s*%s' % self._patterns["varName"], "function": self.__read_var_assignment},
		]

		self.__value_expressions = [
			{"pattern": r'\s*--', "function": self.__read_comment},
			{"pattern": r'\s*(\(|-|not\W|#)', "function": self.__read_formula},
			{"pattern": r'\s*(true|false)\W', "function": self.__read_bool},
			{"pattern": r'\s*nil\W', "function": self.__read_nil},
			{"pattern": r'\s*([.]\d|\d)', "function": self.__read_num},
			{"pattern": r'\s*'+(self._patterns["string"]), "function": self.__read_string},
			{"pattern": r'\s*function\W', "function": self.__read_anonymus_function},
			{"pattern": r'\s*(%s)' % self._patterns["varName2"], "function": self.__read_var},
			{"pattern": r'\s*{', "function": self.__read_table},
		]
	
	def _readMatch(self, match, groupIndex=0):
		self._luaText = self._luaText[match.end(groupIndex):]
		commentMatch = re.match(r'\s*--', self._luaText)
		while commentMatch:
			self.__read_comment()
			commentMatch = re.match(r'\s*--', self._luaText)
	
	def __getVarAttributes(self):
		args = []
		match = re.match(r'\s*([:]|([.])[^.]|\[)', self._luaText)
		if match:
			if match.group(2):
				self._readMatch(match, 2)
			else:
				self._readMatch(match)
		
			nameMatch = None
			if match.group(1) != "[":
				nameMatch = re.match(r'\s*(%s)?' % self._patterns["varName"], self._luaText)
				if nameMatch:
					self._readMatch(nameMatch)
			
			while match:
				if nameMatch and nameMatch.group(1):
					args.append(nameMatch.group(1))
				elif match.group(1) == "[":
					# read arg
					args.append(self._read_value())
					
					# check for ]
					endMatch = re.match(r'\s*\]', self._luaText)
					if endMatch:
						self._readMatch(endMatch)
					else:
						self._raiseSyntaxError("']' expected near %s")
				else:
					self._raiseSyntaxError("<name> expected near %s")
					
				match = re.match(r'\s*([:]|([.])[^.]|\[)', self._luaText)
				if match:
					if match.group(2):
						self._readMatch(match, 2)
					else:
						self._readMatch(match)
				
					nameMatch = None
					if match.group(1) != "[":
						nameMatch = re.match(r'\s*(%s)?' % self._patterns["varName"], self._luaText)
						if nameMatch:
							self._readMatch(nameMatch)

		return args
	
	def _read_expression(self):
		exprRead = False
		expr = None
		for expression in self.__base_expressions:
			if re.match(expression["pattern"], self._luaText):
				expr = expression["function"]()
				exprRead = True
				break
		
		if not exprRead:
			#invalid syntax
			self._raiseSyntaxError("Expected expr in %s")
		else:
			match = re.match(r'\s*;', self._luaText)
			if match:
				self._readMatch(match)
		
		return expr
	
	def _read_value(self):
		valueRead = False
		value = None
		for value_expr in self.__value_expressions:
			if re.match(value_expr["pattern"], self._luaText):
				value = value_expr["function"]()
				valueRead = True
				break

		if not valueRead:
			#invalid syntax
			self._raiseSyntaxError("unexpected symbol near %s")
		else:
			opMatch = re.match(r'\s*(%s)' % self._patterns["operator"], self._luaText)
			if opMatch:
				operator = opMatch.group(1)
				self._readMatch(opMatch)
				value2 = self._read_value()
				return luaFormula(operator, value, value2)
				
		return value
		
	def __read_object_list(self, pattern, isLocal=False):
		objects = {"varNames": [], "args": [], "values": []}
		# set multiple var values at once
		match = re.match(r'\s*', self._luaText)
		while match:
			self._readMatch(match)
			if isinstance(pattern, str):
				match = re.match(pattern, self._luaText)
				if match:
					objects["varNames"].append(match.group(1))
				else:
					self._raiseSyntaxError("unexpected symbol near %s")
				self._readMatch(match)
		
				errorMatch = re.match(r'\s*([.:[])', self._luaText)
				if errorMatch:
					if isLocal:
						self._raiseSyntaxError("unexpected symbol near %s")
					
					objects["args"].append(self.__getVarAttributes())
				else:
					objects["args"].append([])
			else:
				objects["values"].append(pattern())
			match = re.match(r'\s*,', self._luaText)
		
		return objects
	
	def __read_comment(self):
		match = re.match(r'\s*--(\[(=)*\[)?', self._luaText)
		
		if match.group(1):
			commentMatch = re.match(r'\s*--\[(=*)\[((:?.|\n)*?)\]\1\]', self._luaText)
		else:
			commentMatch = re.match(r'\s*--.*?\n', self._luaText)
		
		self._readMatch(commentMatch)

	def __read_formula(self):
		match = re.match(r'\s*(\(|-|not|#)', self._luaText)
		self._readMatch(match)
		
		value = self._read_value()
		
		if match.group(1) == "(":
			endMatch = re.match(r'\s*\)', self._luaText)
			if endMatch:
				self._readMatch(endMatch)
			else:
				self._raiseSyntaxError("Expected ')' near %s")
			
		return luaFormula(match.group(1), value)

	def __read_nil(self):
		match = re.match(r'\s*nil', self._luaText)
		self._readMatch(match)

		return None

	def __read_bool(self):
		match = re.match(r'\s*(true|false)\W', self._luaText)
		value = "true" in match.group(1)
		self._readMatch(match, 1)

		return value

	def __read_num(self):
		match = re.match(r'\s*((\d+[.]?\d*|\d*[.]?\d+)(e[-]?\d+)?)', self._luaText)
		
		num = toNum(match.group(1))
		self._readMatch(match)

		return num

	def __read_string(self):
		match = re.match(r'\s*%s' % (self._patterns["string"]), self._luaText)
		if match.group(1) == '"':
			endPattern = r'\n|([^\\]|^)"'
		elif match.group(1) == "'":
			endPattern = r"\n|([^\\]|^)'"
		else:
			endPattern = r'([^\\]|^)\]%s\]' % match.group(2)
		self._luaText = self._luaText[match.end():]
		
		endMatch = re.search(endPattern, self._luaText)
		if endMatch:
			if endMatch.group(0) == "\n":
				self._luaText = match.group(1) + self._luaText
				self._raiseSyntaxError("unfinished string near %s", r'(.*)\n')
			str = self._luaText[:endMatch.start()+len(endMatch.group(1))]
			self._readMatch(endMatch)
		else:
			self._luaText = ""
			self._raiseSyntaxError("unfinished string near %s")

		str = re.sub(r'\\[\\"\'abfnrtv]', lambda match: codecs.decode(match.group(0), 'unicode-escape'), str)

		return str

	def __read_anonymus_function(self):
		match = re.match(r'(\s*function)\W', self._luaText)
		self._readMatch(match, 1)
		match = re.match(r'\s*\(', self._luaText)
		if match:
			self._readMatch(match)
		else:
			self._raiseSyntaxError("'(' expected near %s")
		
		argNames = []
		match = re.match(r'\s*(%s)' % self._patterns["varName2"], self._luaText)
		while match:
			if match:
				argName = match.group(1)
			else:
				self._raiseSyntaxError("unexpected symbol near %s")
			self._readMatch(match)
			
			argNames.append(argName)
			
			if argName == "...":
				break
	
			errorMatch = re.match(r'\s*([.:[])', self._luaText)
			if errorMatch:
				self._raiseSyntaxError("unexpected symbol near %s")
			match = re.match(r'\s*,', self._luaText)
			if match:
				self._readMatch(match)
				match = re.match(r'\s*(%s)' % self._patterns["varName2"], self._luaText)
				if not match:
					self._raiseSyntaxError("<name> or '...' expected near %s")
		
		match = re.match(r'\s*\)', self._luaText)
		if match:
			self._readMatch(match)
		else:
			self._raiseSyntaxError("')' expeted near %s")
		
		function = luaFunction(argNames, self._luaText, self)
		self._luaText = function.read_expressions(r'\s*end(\W|$)')
		endMatch = re.match(r'(\s*end)(\W|$)', self._luaText)
		if endMatch:
			self._readMatch(endMatch, 1)
		else:
			self._raiseSyntaxError("'end' expeted near %s")
		
		return function

	def __read_function_call(self, varName, args=[]):		
		funcArgs = []
		match = re.match(r'\s*\(', self._luaText)
		if match:
			self._readMatch(match)
			endMatch = re.match(r'\s*\)', self._luaText)
			while not endMatch:
				arg = self._read_value()
				funcArgs.append(arg)
				
				# check for end of argument list
				endMatch = re.match(r'\s*\)', self._luaText)
				if not endMatch:
					match = re.match(r'\s*,', self._luaText)
					if not match:
						self._raiseSyntaxError("Expected ')' near %s")
					else:
						self._readMatch(match)
						
			if endMatch:
				self._readMatch(endMatch)

		else:
			arg = self.__read_string()
			funcArgs.append(arg)

		args2 = []
		argMatch = re.match(r'\s*([:[]|[.][^.])', self._luaText)
		if argMatch:			
			args2 = self.__getVarAttributes()

		return luaFunctionCall(luaVariable(varName, args), funcArgs, args2)

	def __read_var(self):
		match = re.match(r'\s*(%s)' % self._patterns["varName2"], self._luaText)
		args = []
		varName = match.group(1)
		if varName in _reserved:
			self._raiseSyntaxError("unexpected symbol near %s")
		value = None

		self._readMatch(match)
		
		argMatch = re.match(r'\s*([:[]|[.][^.])', self._luaText)
		if argMatch:
			args = self.__getVarAttributes()
			
		funcMatch = re.match(r'\s*(\(|%s)' % self._patterns["string"], self._luaText)
		if funcMatch:
			return self.__read_function_call(varName, args)

		return luaVariable(varName, args)

	def __read_table(self):
		match = re.match(r'\s*{', self._luaText)
		self._readMatch(match)
		
		table = luaTable()
		endMatch = re.match(r'\s*}', self._luaText)
		while not endMatch:
			# read table entries
			# get key
			match = re.match(r'\s*(\[|(%s))' % self._patterns["varName"], self._luaText)
			
			key = None
			if match:
				if "[" in match.group(1):
					self._readMatch(match)
					key = self._read_value()
					
					match = re.match(r'\s*\]', self._luaText)
					if not match:
						self._raiseSyntaxError("']' expected near %s")
					self._readMatch(match)
					
					match = re.match(r'\s*=', self._luaText)
					if not match:
						self._raiseSyntaxError("'=' expected near %s")
					self._readMatch(match)
					
					if key == None:
						self._raiseSyntaxError("table index is nil")
				else:
					key = match.group(1)
					textBackup = self._luaText
					self._readMatch(match)
					match = re.match(r'\s*=', self._luaText)
					if match:
						self._readMatch(match)
					else:
						self._luaText = textBackup
						key = None
			
			# get value
			value = self._read_value()
			
			if key:
				table.add_key_value(value, key)
			else:
				table.add_value(value)

			match = re.match(r'\s*,', self._luaText)
			if match:
				self._readMatch(match)
			
			# check for end of table
			endMatch = re.match(r'\s*}', self._luaText)
			if not endMatch and not match:
				break
					
		if endMatch:
			self._readMatch(endMatch)
		else:
			self._raiseSyntaxError("'}' expected near %s")

		return table

	def __read_return(self):
		match = re.match(r'(\s*return)\W', self._luaText)
		self._readMatch(match, 1)
		
		objects = self.__read_object_list(self._read_value)
		values = objects["values"]
		return luaExpression("return", values)

	def __read_break(self):
		if self._insideLoop:
			match = re.match(r'(\s*break)\W', self._luaText)
			self._readMatch(match, 1)
			return luaExpression("break")
		else:
			self._raiseSyntaxError("<break> not inside a loop", r'')

	def __read_for_loop(self):
		match = re.match(r'(\s*for)\W', self._luaText)
		self._readMatch(match, 1)
		
		match = re.match(r'\s*(%s)' % self._patterns["varName2"], self._luaText)
		if match:
			self._readMatch(match)
			
			vars = [match.group(1)]
			match = re.match(r'\s*(,|=|in\W)', self._luaText)
			if match:
				if match.group(1) != "=":
					# iterator based loop
					listMatch = re.match(r'\s*,', self._luaText)
					while listMatch:
						self._readMatch(listMatch)
						nameMatch = re.match(r'\s*(%s)' % self._patterns["varName2"], self._luaText)
						if nameMatch:
							vars.append(nameMatch.group(1))
							self._readMatch(nameMatch)
						else:
							self._raiseSyntaxError("<name> expected near %s")
						
						listMatch = re.match(r'\s*,', self._luaText)
					
					match = re.match(r'(\s*in)\W', self._luaText)
					if match:
						self._readMatch(match, 1)
						iterator = self._read_value()
					else:
						self._raiseSyntaxError("'in' expected near %s")
					
				else:
					# start end based loop
					# get start and end
					self._readMatch(match)
					start = self._read_value()
					match = re.match(r'\s*,', self._luaText)
					if match:
						self._readMatch(match)
						end = self._read_value()
					else:
						self._raiseSyntaxError("',' expected near %s")
					
					# optional step argument
					step = 1
					match = re.match(r'\s*,', self._luaText)
					if match:
						self._readMatch(match)
						step = self._read_value()
					
					# create iterator
					iterator = luaIterator(start, end, step)
			
			else:
				self._raiseSyntaxError("'=' or 'in' expected near %s")
			
			match = re.match(r'(\s*do)\W', self._luaText)
			if match:
				self._readMatch(match, 1)
				forLoop = luaForLoop(vars, iterator, self._luaText, self)
				self._luaText = forLoop.read_expressions(r'\s*end(\W|$)')
				endMatch = re.match(r'(\s*end)(\W|$)', self._luaText)
				if endMatch:
					self._readMatch(endMatch, 1)
				else:
					self._raiseSyntaxError("'end' expeted near %s")
				return luaExpression("control_structure", forLoop)
			else:
				self._raiseSyntaxError("'do' expected near %s")
		else:
			self._raiseSyntaxError("<name> expected near %s")

	def __read_if_statement(self):
		match = re.match(r'(\s*if)\W', self._luaText)
		self._readMatch(match, 1)
		
		condition = self._read_value()
		match = re.match(r'(\s*then)\W', self._luaText)
		if match:
			self._readMatch(match, 1)
			ifStatement = luaIfStatement(condition, self._luaText, self)
			self._luaText = ifStatement.read_expressions()
			return luaExpression("control_structure", ifStatement)			
		else:
			self._raiseSyntaxError("'then' expected near %s")

	def __read_while_loop(self):
		match = re.match(r'(\s*while)\W', self._luaText)
		self._readMatch(match, 1)
		
		condition = self._read_value()
		match = re.match(r'(\s*do)\W', self._luaText)
		if match:
			self._readMatch(match, 1)
			whileLoop = luaWhileLoop(condition, self._luaText, self)
			self._luaText = whileLoop.read_expressions(r'\s*end(\W|$)')
			endMatch = re.match(r'(\s*end)(\W|$)', self._luaText)
			if endMatch:
				self._readMatch(endMatch, 1)
				return luaExpression("control_structure", whileLoop)	
			else:
				self._raiseSyntaxError("'end' expeted near %s")
		else:
			self._raiseSyntaxError("'do' expected near %s")

	def __read_repeat_loop(self):
		match = re.match(r'(\s*repeat)\W', self._luaText)
		self._readMatch(match, 1)
		
		repeatLoop = luaRepeatLoop(self._luaText, self)
		self._luaText = repeatLoop.read_expressions(r'\s*until(\W)')
		endMatch = re.match(r'(\s*until)\W', self._luaText)
		if endMatch:
			self._readMatch(endMatch, 1)
		else:
			self._raiseSyntaxError("'until' expeted near %s")
		
		condition = self._read_value()
		repeatLoop._addCondition(condition)
		return luaExpression("control_structure", repeatLoop)

	def __read_do_block(self):
		match = re.match(r'(\s*do)\W', self._luaText)
		self._readMatch(match, 1)
		
		doBlock = luaDoBlock(self._luaText, self)
		self._luaText = doBlock.read_expressions(r'(\s*end)(\W|$)')
		endMatch = re.match(r'(\s*end)(\W|$)', self._luaText)
		if endMatch:
			self._readMatch(endMatch, 1)
		else:
			self._raiseSyntaxError("'end' expeted near %s")

		return luaExpression("control_structure", doBlock)

	def __read_local(self):
		match = re.match(r'\s*(local)\W', self._luaText)
		isLocal = (match.group(1) != None)
		self._readMatch(match, 1)
		
		match = re.match(r'\s*function\W', self._luaText)
		if match:
			return self.__read_named_function(isLocal)
		else:
			return self.__read_var_assignment(isLocal)

	def __read_var_assignment(self, isLocal=False):		
		values = []
		objects = self.__read_object_list(r'\s*(%s)' % self._patterns["varName"], isLocal)
		varNames = objects["varNames"]
		args = objects["args"]
		
		match = re.match(r'\s*(=|\(|%s)' % self._patterns["string"], self._luaText)
		if match:
			if match.group(1) == "=":
				self._readMatch(match)
				objects = self.__read_object_list(self._read_value, isLocal)
				values = objects["values"]
			else:
				if isLocal:
					self._raiseSyntaxError("syntax error near %s")
				elif len(varNames) > 1:
					self._raiseSyntaxError("syntax error near %s")
				else:
					return self.__read_function_call_expr(varNames[0], args[0])
				
		else:
			if not isLocal:
				self._raiseSyntaxError("syntax error near %s")

		return luaExpression("assign_var", values, varNames, isLocal, args)
		
	def __read_named_function(self, isLocal=False):		
		match = re.match(r'\s*(function)\W', self._luaText)
		self._readMatch(match, 1)
		
		match = re.match(r'\s*(%s)' % self._patterns["varName"], self._luaText)
		self._readMatch(match)
		args = []
		varName = match.group(1)
		value = None
		
		errorMatch = re.match(r'\s*([.:[])', self._luaText)
		if errorMatch:
			if isLocal:
				self._raiseSyntaxError("unexpected symbol near %s", r'\s*([.:[])')
			
			args = self.__getVarAttributes()
		
		self._luaText = "function " + self._luaText
		value = self._read_value()

		return luaExpression("assign_var", [value], [varName], isLocal, [args])

	def __read_function_call_expr(self, varName, args):
		value = self.__read_function_call(varName, args)
		return luaExpression("function_call", value)
	
	def __getBaseScriptInfo(self):
		if isinstance(self._parent, luaControlStructure) and not isinstance(self._parent, luaParser):
			return self._parent.__getBaseScriptInfo()
		else:
			return self.__fullText, self.__filename
		
	def _raiseSyntaxError(self, message, pattern = r'\s*(\S+)'):
		fullText, fileName = self.__getBaseScriptInfo()
		index = fullText.find(self._luaText)
		if self._luaText == "":
			line = fullText.count("\n")+1
			errorText = fullText.rpartition("\n")[2]
			offset = len(errorText)+1
			
			if "%s" in message:
				errorMessage = message % "<eof>"
			else:
				errorMessage = message
		else:
			if index > -1:
				line = fullText.count("\n", 0, index)+1
				lineStart = fullText.rfind("\n", 0, index)
				offset = index - lineStart
			else:
				line = 0
				offset = 0
			
			errorText = re.sub(r'.*\n', "", fullText, line-1)
			errorText = errorText.partition("\n")[0]
			
			match = re.match(pattern, self._luaText)
			if match:
				matchText = match.group(1)
				matchText = re.sub(r'(\S\w*).*', r'\1', matchText)
			else:
				matchText = self._luaText
			
			if "%s" in message:
				errorMessage = message % ("'%s'" % matchText)
			else:
				errorMessage = message
			
		raise SyntaxError(errorMessage, (fileName, line, offset, errorText))
	
	def _execute_expressions(self, varScope, exprList = None):
		if exprList == None:
			exprList = self._expressions
		for expr in exprList:
			if expr.type == "return":
				return luaReturn(expr.execute(varScope))
			elif expr.type == "break":
				return luaBreak()
			else:
				result = expr.execute(varScope)
				if isinstance(result, (luaBreak, luaReturn)):
					return result
		return None
	
	def read_expressions(self, endPattern, exprList = None):
		if exprList == None:
			exprList = self._expressions

		match = re.match(endPattern+r'|\s*$', self._luaText)
		while not match:
			expr = self._read_expression()
			if expr:
				exprList.append(expr)
			match = re.match(endPattern+r'|\s*$', self._luaText)
		
		text = self._luaText
		self._luaText = ""
		return text
		
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild()
		else:
			varScope = luaVarScope()
		
		# execute expression
		for expr in self._expressions:
			if expr.type == "return":
				result = expr.execute(varScope)
				del varScope
				return result
			else:
				result = expr.execute(varScope)
				if isinstance(result, luaReturn):
					del varScope
					result = result.getValue()
					
					if isinstance(result, (list, tuple)) and len(result) == 1:
						result = result[0]
					
					return result
	
	def execute(self, parentVarScope):
		return self._run(parentVarScope)

class luaFunction(luaControlStructure, luaValue):
	def __init__(self, argNames, text, parent):
		luaControlStructure.__init__(self, text, parent)
		self.__argNames = argNames
		
	def _run(self, parentVarScope, args):
		# map given arguments to argument names
		vars = {}
		for i in range(len(self.__argNames)):
			if i < len(args):
				if self.__argNames[i] == "...":
					vars[self.__argNames[i]] = list(args[i:])
					break
				else:
					if isinstance(args[i], list):
						# convert list to dict
						arg = dict(zip(range(1, len(args[i])+1), args[i]))
					else:
						arg = args[i]
					vars[self.__argNames[i]] = arg
			else:
				vars[self.__argNames[i]] = None

		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild(vars)
		else:
			varScope = luaVarScope(vars)

		# execute expressions
		result = self._execute_expressions(varScope)
		del varScope
		if isinstance(result, luaReturn):
			result = result.getValue()
			
			if isinstance(result, (list, tuple)) and len(result) == 1:
				result = result[0]
			
			return result
		
	def getValue(self, varScope):
		return lambda *args: self._run(varScope, argumentValidation(args))

class luaForLoop(luaControlStructure):
	def __init__(self, varNames, iterator, text, parent):
		luaControlStructure.__init__(self, text, parent)
		self._insideLoop = True
		self.__varNames = varNames
		self.__iterator = iterator
		
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild()
		else:
			varScope = luaVarScope()

		iterator = self.__iterator
		if isinstance(iterator, luaValue):
			iterator = iterator.getValue(varScope)
		
		for item in iterator:
			# map given arguments to argument names
			vars = {}
			if isinstance(item, (tuple, list)):
				for i in range(len(self.__varNames)):
					if i < len(item):
						vars[self.__varNames[i]] = item[i]
					else:
						vars[self.__varNames[i]] = None
			else:
				vars[self.__varNames[0]] = item
			
			# reset var scope
			varScope.reset(vars)
			
			# execute expressions
			result = self._execute_expressions(varScope)
			if isinstance(result, luaBreak):
				del varScope
				return None
			elif isinstance(result, luaReturn):
				del varScope
				return result

class luaIfStatement(luaControlStructure):
	def __init__(self, condition, text, parent):
		luaControlStructure.__init__(self, text, parent)
		self.__condition = condition
		self.__elifs = []
		self.__else = []
	
	def read_expressions(self):
		# read if block
		self._luaText = luaControlStructure.read_expressions(self, r'\s*(elseif|else|end)(\W|$)')
		
		# read elseif blocks (optional)
		elifMatch = re.match(r'(\s*elseif)\W', self._luaText)
		while elifMatch:
			self._readMatch(elifMatch, 1)
			condition = luaControlStructure._read_value(self)
			match = re.match(r'(\s*then)\W', self._luaText)
			if match:
				self._readMatch(match, 1)
				elifBlock = {"condition": condition, "expr": []}
				self._luaText = luaControlStructure.read_expressions(self, r'\s*(elseif|else|end)(\W|$)', elifBlock["expr"])
				self.__elifs.append(elifBlock)
				elifMatch = re.match(r'(\s*elseif)\W', self._luaText)
			else:
				self._raiseSyntaxError("then expected near %s")
		
		# read else block (optional)
		elseMatch = re.match(r'(\s*else)\W', self._luaText)
		if elseMatch:
			self._readMatch(elseMatch, 1)
			self._luaText = luaControlStructure.read_expressions(self, r'\s*end(\W|$)', self.__else)

		endMatch = re.match(r'(\s*end)(\W|$)', self._luaText)
		if endMatch:
			self._readMatch(endMatch, 1)
		else:
			self._raiseSyntaxError("'end' expected near %s")
		
		text = self._luaText
		self._luaText = ""
		return text
	
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild()
		else:
			varScope = luaVarScope()

		condition = self.__condition
		if isinstance(condition, luaValue):
			condition = condition.getValue(varScope)
		condition = toBool(condition)
		
		if condition:
			result = self._execute_expressions(varScope)
			del varScope
			return result
		else:
			elifTrue = False
			for elifBlock in self.__elifs:
				condition = elifBlock["condition"]
				if isinstance(condition, luaValue):
					condition = condition.getValue(varScope)
				condition = toBool(condition)
				
				if condition:
					elifTrue = True
					result = self._execute_expressions(varScope, elifBlock["expr"])
					del varScope
					return result
			if not elifTrue:
				result = self._execute_expressions(varScope, self.__else)
				del varScope
				return result

class luaWhileLoop(luaControlStructure):
	def __init__(self, condition, text, parent):
		luaControlStructure.__init__(self, text, parent)
		self._insideLoop = True
		self.__condition = condition
		
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild()
		else:
			varScope = luaVarScope()

		condition = self.__condition
		if isinstance(condition, luaValue):
			condition = condition.getValue(varScope)
		condition = toBool(condition)
			
		while condition:
			# reset var scope
			varScope.reset()
			
			# execute expressions
			result = self._execute_expressions(varScope)
			if isinstance(result, luaBreak):
				del varScope
				return None
			elif isinstance(result, luaReturn):
				del varScope
				return result
			
			condition = self.__condition
			if isinstance(condition, luaValue):
				condition = condition.getValue(varScope)
			condition = toBool(condition)

class luaRepeatLoop(luaControlStructure):
	def __init__(self, text, parent):
		luaControlStructure.__init__(self, text, parent)
		self._insideLoop = True
		self.__condition = True
	
	def _addCondition(self, condition):
		self.__condition = condition
		
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild()
		else:
			varScope = luaVarScope()

		condition = False			
		while not condition:
			# reset var scope
			varScope.reset()
			
			# execute expressions
			result = self._execute_expressions(varScope)
			if isinstance(result, luaBreak):
				del varScope
				return None
			elif isinstance(result, luaReturn):
				del varScope
				return result
			
			condition = self.__condition
			if isinstance(condition, luaValue):
				condition = condition.getValue(varScope)
			condition = toBool(condition)

class luaDoBlock(luaControlStructure):
	def __init__(self, text, parent):
		luaControlStructure.__init__(self, text, parent)
		
	def _run(self, parentVarScope):
		# create var scope
		if isinstance(parentVarScope, luaVarScope):
			varScope = parentVarScope.getChild(vars)
		else:
			varScope = luaVarScope(vars)
		
		# execute expressions
		result = self._execute_expressions(varScope)
		return result
		del varScope

class luaParser(luaControlStructure):
	def __init__(self):
		luaControlStructure.__init__(self, "", None)
		# global standard vars
		vars = globalVars
		vars["_"] = lambda args: argumentValidation(args)[0]
		vars["require"] = lambda args: self.__require(argumentValidation(args)[0])
		vars["table"] = lua_table.table
		vars["math"] = lua_math.maths
		vars["package"] = {
			"loaded": {}
		}
		self._varScope = luaVarScope(vars)
		self.__loadedFiles = {}
		
	def __del__(self):
		del self._varScope
	
	def __require(self, name):
		# to do: error message and actual function
		if not isinstance(name, (str, int, float)):
			raise ValueError("bad argument #1 to 'require' (string expected, got %s)" % luaType(name))
		else:
			if isinstance(name, (int, float)):
				name = str(name)
			package = self._varScope.getVar("package")
			if name in package["loaded"]:
				return self.executeScript(package["loaded"][name])
			else:
				if "path" in package:
					searchPaths = package["path"].split(";")
					searchedPaths = ""
					for path in searchPaths:
						path = path.replace("?", name)
						if isfile(path):
							script = self.loadScriptFile(path)
							package["loaded"][name] = script
							result = self.executeScript(script)
							return result
						else:
							searchedPaths += "\n      no file '%s'" % path
					raise RuntimeError("module '%s' not found:%s" % (name, searchedPaths))
				else:
					raise RuntimeError("module '%s' not found" % name)
		
	def setVar(self, varName, value, args=[]):
		if isinstance(value, list):
			# convert list to dict
			value = dict(zip(range(1, len(value)+1), value))
		self._varScope.setVar(varName, value, False, args)
	
	def getVar(self, varName, args=[]):
		return self._varScope.getVar(varName, args)
	
	def showVars(self):
		self._varScope.showVars()
	
	def loadScriptFile(self, filename, varScope=None):
		if varScope == None:
			varScope = self._varScope
		
		if filename in self.__loadedFiles:
			return self.__loadedFiles[filename]
		else:
			if isfile(filename):
				scriptFile = open(filename, encoding='utf-8', mode='r')
			
			if scriptFile:
				text = scriptFile.read()
				if text[0] == '\ufeff':
					text = text[1:]
				scriptFile.close()
				script = luaControlStructure(text, self, filename)
				script.read_expressions("$")
				self.__loadedFiles[filename] = script
				
				return script
		
		return None
	
	def executeScript(self, script, varScope=None):
		if varScope == None:
			varScope = self._varScope
		
		if isinstance(script, luaControlStructure):
			result = script.execute(varScope)
			
			return result
		