import bpy
from mathutils import *
from math import *
from ..tf_util.base_functions import TF_base_functions, TFPropertyGroup
from ..tf_util.mesh_util import TFMeshSetting
from ..tf_util.model_util import TFLodSetting

plugin_info = {
	"name": "Custom Metadata Provider",
	"author": "eis_os, Merk",
	"version": "0.7.0",
	"order": 10,
}

_tfproperty_types = [
		("prop_string", "String", "", 0),
		("prop_int", "Integer", "", 1),
		("prop_float", "Float", "", 2),
		("prop_bool", "Boolean", "", 3),
		("prop_table", "Table", "", 4),
		("empty", "Empty", "", 6)
]

_engine_list = [
	["Steam", "STEAM", "", 0],
	["Diesel", "DIESEL", "", 1],
	["Electric", "ELECTRIC", "", 2],
	["Horse", "HORSE", "", 3],
]

_cargo_type_list_enum = [
	["Passengers", "PASSENGERS", ""],
	["Goods", "GOODS", ""],
	["Coal", "COAL", ""],
	["Iron Ore", "IRON_ORE", ""],
	["Wood", "WOOD", ""],
	["Oil", "OIL", ""],
	["Logs", "LOGS", ""],
	["Livestock", "LIVESTOCK", ""],
	["Stone", "STONE", ""],
	["Grain", "GRAIN", ""],
	["Crude oil", "CRUDE", ""],
	["Steel", "STEEL", ""],
	["Planks", "PLANKS", ""],
	["Plastic", "PLASTIC", ""],
	["Slag", "SLAG", ""],
	["Construction material", "CONSTRUCTION_MATERIALS", ""],
	["Machines", "MACHINES", ""],
	["Fuel", "FUEL", ""],
	["Tools", "TOOLS", ""],
	["Food", "FOOD", ""],
]

# create additional lists for faster access
_engine_types = []
for engine_type in _engine_list:
	_engine_types.append(engine_type[1])

_cargo_types = []
for cargo_type in _cargo_type_list_enum:
	_cargo_types.append(cargo_type[1])

_carrier_list = [
	["NONE", "NONE", "", 3, {
		"vehicletype": "",
		"keys": [
			"soundSet",
		],
		"configs": [],
	}],
	["RAIL", "RAIL", "", 0, {
		"vehicletype": "railVehicle",
		"keys": [
			"engines",
			"soundSet",
			{"prop_name": "topSpeed", "text": "Topspeed"},
			{"prop_name": "weight", "text": "Weight"},
		],
		"configs": ["axles", "backBackwardParts", "backForwardParts", "blinkingLights0", "blinkingLights1", "fakeBogies", "frontBackwardParts", "frontForwardParts", "innerBackwardParts", "innerForwardParts", "wheels"],
	}],
	["ROAD", "ROAD", "", 1, {
		"vehicletype": "roadVehicle",
		"keys": [
			"soundSet",
			{"prop_name": "topSpeed", "text": "Topspeed"},
			{"prop_name": "weight", "text": "Weight"},
			{"prop_name": "power", "text": "Power"},
			{"prop_name": "tractiveEffort", "text": "Tractive Effort"},
		],
		"configs": ["axles", "blinkLightsLeft", "blinkLightsRight", "brakeLights", "fakeBogies", "headLights", "steeringParts", "wheels"],
	}],
	["TRAM", "TRAM", "", 2, {
		"vehicletype": "railVehicle",
		"keys": [
			"engines",
			"soundSet",
			{"prop_name": "topSpeed", "text": "Topspeed"},
			{"prop_name": "weight", "text": "Weight"},
		],
		"configs": ["axles", "backBackwardParts", "backForwardParts", "blinkingLights0", "blinkingLights1", "fakeBogies", "frontBackwardParts", "frontForwardParts", "innerBackwardParts", "innerForwardParts", "wheels"],
	}],
	["WATER", "WATER", "", 4, {
		"vehicletype": "waterVehicle",
		"keys": [
			"soundSet",
			{"prop_name": "topSpeed", "text": "Topspeed"},
			{"prop_name": "weight", "text": "Weight"},
			{"prop_name": "power", "name": "availPower", "text": "available Power"},
			{"prop_name": "maxRpm", "text": "max. RPM"},
			{"prop_name": "area", "text": "Area"},
		],
		"configs": ["paddels", "rudder", "flags"],
	}],
	["AIR", "AIR", "", 5, {
		"vehicletype": "airVehicle",
		"keys": [
			"soundSet",
			{"prop_name": "topSpeed", "text": "Topspeed"},
			{"prop_name": "weight", "text": "Weight"},
			{"prop_name": "power", "name": "maxThrust", "text": "max. Thrust"},
			{"prop_name": "tractiveEffort", "name": "idleThrust", "text": "idle Thrust"},
			{"prop_name": "maxPayload", "text": "max. Payload"},
			{"prop_name": "maxTakeOffWeight", "text": "max. Takeoff Weight"},
			{"prop_name": "timeToFullThrust", "text": "Time to full Thrust"},
			{"prop_name": "area", "name": "wingArea", "text": "Wing Area"},
		],
		"configs": ["axles", "axleRadii", "beaconlights", "fakeBogies", " landingLight", "wheels", "elevator", "aileronLeft", "aileronRight", "flaps", "props", "propsB", "rudder", "steeringParts", "strobeLights", "sheelRadii"],
	}],
]

_mesh_config_keys = [
	["axles", "Axle"],
	["backBackwardParts", "Back backward parts"],
	["backForwardParts", "Back forward parts"],
	["blinkingLights0", "Blinking lights 0"],
	["blinkingLights1", "Blinking lights 1"],
	["fakeBogies", "Fake bogies"],
	["frontBackwardParts", "Front backward parts"],
	["frontForwardParts", "Front forward parts"],
	["innerBackwardParts", "Inner backward parts"],
	["innerForwardParts", "Inner forward parts"],
	["wheels", "Wheel"],
]

nameList = []

def _get_children_enum(self, context):
	childListEnum = []
	if hasattr(self, "id_data"):
		childList = TF_base_functions.get_children(self.id_data)
		for child in childList:
			if child.name not in nameList:
				nameList.append(child.name)
				index = len(nameList)-1
			else:
				index = nameList.index(child.name)
			childListEnum.append((nameList[index], nameList[index], nameList[index]))

	return childListEnum

class TFLodConfigIdProp(TFPropertyGroup):
	id = bpy.props.EnumProperty(items = _get_children_enum)
	#bpy.props.StringProperty()

class TFLodConfigPropGroup(TFPropertyGroup):
	prop_names = ["ids"]
	ids = bpy.props.CollectionProperty(type = TFLodConfigIdProp)
	ids_activeIndex = bpy.props.IntProperty()

class TFLodConfigIdPropAddOp(bpy.types.Operator):
	bl_label = "Add Id"
	bl_idname = "collection.tflodconfigidpropgroup_add"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		obj = context.object
		if hasattr(obj, "tf_lod"):
			metadata = obj.tf_lod.metadata
		
		if hasattr(metadata,  self.attr):
			prop = getattr(metadata, self.attr)
			collection = prop.ids
		
			item = collection.add()
			prop.ids_activeIndex = len(collection)-1
		return {'FINISHED'}

class TFLodConfigIdPropRemoveOp(bpy.types.Operator):
	bl_label = "Remove Id"
	bl_idname = "collection.tflodconfigidpropgroup_remove"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		obj = context.object
		if hasattr(obj, "tf_lod"):
			metadata = obj.tf_lod.metadata
		
		if hasattr(metadata, self.attr):
			prop = getattr(metadata, self.attr)
			collection = prop.ids
			index = prop.ids_activeIndex
			collection.remove(index)
			if index > 0:
				prop.ids_activeIndex = index-1
		return {'FINISHED'}

class TFLodConfigIdPropUIList(bpy.types.UIList):
	def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
		layout.prop(item, "id", "")

def __drawLodConfigPropGroup(boxKey, mydata, key):
	row = boxKey.row()
	row.template_list(
		listtype_name="TFLodConfigIdPropUIList",
		dataptr=mydata,
		propname="ids",
		active_dataptr=mydata,
		active_propname="ids_activeIndex",
		rows = 1,
		maxrows = 2,
		list_id = "TFLodConfigIdPropUIList_"+key,
	)
	col = row.column(align=True)
	col.operator("collection.tflodconfigidpropgroup_add", icon="ZOOMIN", text="").attr = key
	col.operator("collection.tflodconfigidpropgroup_remove", icon="ZOOMOUT", text="").attr = key

class TFLodConfigPropGroupMaxAngle(TFPropertyGroup):
	prop_names = ["ids", "maxAngle"]
	ids = bpy.props.CollectionProperty(type = TFLodConfigIdProp)
	ids_activeIndex = bpy.props.IntProperty()
	maxAngle = bpy.props.FloatProperty(default=30.0, step=100)

def __drawLodConfigPropGroupMaxAngle(boxKey, mydata, key):
	row = boxKey.row()
	row.template_list(
		listtype_name="TFLodConfigIdPropUIList",
		dataptr=mydata,
		propname="ids",
		active_dataptr=mydata,
		active_propname="ids_activeIndex",
		rows = 1,
		maxrows = 2,
		list_id = "TFLodConfigIdPropUIList_"+key,
	)
	col = row.column(align=True)
	col.operator("collection.tflodconfigidpropgroup_add", icon="ZOOMIN", text="").attr = key
	col.operator("collection.tflodconfigidpropgroup_remove", icon="ZOOMOUT", text="").attr = key
	boxKey.prop(mydata, "maxAngle", text="max. Angle")

def __TFLodConfigPropGroupSetValue(data, object, idToObject):
	if hasattr(object, "ids"):
		for meshID in data["ids"]:
			if meshID <= len(idToObject):
				id = object.ids.add()
				id.id = idToObject[meshID-1][0].name
	
	if "maxAngle" in data and hasattr(object, "maxAngle"):
		object.maxAngle = data["maxAngle"] 

def __TFLodConfigPropGroupGetValue(data, children):
	result = {"ids": []}
	if hasattr(data, "ids"):
		for id in data.ids:
			if id.id in bpy.data.objects and bpy.data.objects[id.id] in children:
				result["ids"].append(children.index(bpy.data.objects[id.id])+1)
		
	if len(result["ids"]) == 0:
		return None
	
	if hasattr(data, "maxAngle"):
		result["maxAngle"] = data.maxAngle

	return result

_lod_config_classes = [
	TFLodConfigIdProp,
	TFLodConfigPropGroup,
	TFLodConfigIdPropAddOp,
	TFLodConfigIdPropRemoveOp,
	TFLodConfigIdPropUIList,
	TFLodConfigPropGroupMaxAngle,
]

_lod_config_keys = [
	["paddles", "Paddles", TFLodConfigPropGroup, __drawLodConfigPropGroup, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["rudder", "Rudder", TFLodConfigPropGroupMaxAngle, __drawLodConfigPropGroupMaxAngle, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["flags", "Flags", TFLodConfigPropGroup, __drawLodConfigPropGroup, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["elevator", "Elevator", TFLodConfigPropGroupMaxAngle, __drawLodConfigPropGroupMaxAngle, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["aileronLeft", "Aileron left", TFLodConfigPropGroupMaxAngle, __drawLodConfigPropGroupMaxAngle, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["aileronRight", "Aileron right", TFLodConfigPropGroupMaxAngle, __drawLodConfigPropGroupMaxAngle, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["flaps", "Flaps", TFLodConfigPropGroupMaxAngle, __drawLodConfigPropGroupMaxAngle, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
	["props", "Propellers", TFLodConfigPropGroup, __drawLodConfigPropGroup, __TFLodConfigPropGroupSetValue, __TFLodConfigPropGroupGetValue],
]

# create additional lists for faster access
_cargo_type_list = []
for cargo_type in _cargo_type_list_enum:
	_cargo_type_list.append(cargo_type[0])

_carrier_info = {}
_engine_props = []
for carrier in _carrier_list:
	if len(carrier) > 4:
		_carrier_info[carrier[0]] = carrier[4]

def get_basetype(vartype):
	if vartype in {"<class 'int'>", "<class 'long'>", "<class 'float'>"}:
		return "num"
	elif vartype == "<class 'str'>":
		return "string"
	elif vartype == "<class 'bool'>":
		return "bool"
	elif vartype in {"<class 'dict'>", "<class 'list'>"}:
		return "table"
	else:
		return "unknown"

def draw_customMetadata(layout, mydata, basetype):
	boxCustomMeta = TF_base_functions.draw_hideable_box(layout, "Custom Metadata:", basetype + "_custom_metadata")
	if boxCustomMeta:
		row = boxCustomMeta.row()
		row.template_list(
			listtype_name="TFCustomMetadataSettingsItemUIList",
			dataptr=mydata,
			propname="customMetadata",
			active_dataptr=mydata,
			active_propname="customMetadata_activeidx",
			rows = 1,
			maxrows = 3,
			list_id = "TFCustomMetadataSettingsItemUIList_"+basetype,
		)
		col = row.column(align=True)
		col.operator("collection.tfcustommetadata_add", icon="ZOOMIN", text="").attr = mydata.customMetadata.path_from_id()
		col.operator("collection.tfcustommetadata_remove", icon="ZOOMOUT", text="").attr = mydata.customMetadata.path_from_id()

# Metadata
class TFMetadataGroup(TFPropertyGroup):
	prop_names = []

	def get_props(self):
		props = self.prop_names.copy()
		if hasattr(self, "customMetadata"):
			i = 0
			for prop in self.customMetadata:
				props.append({"prop_name": "customMetadata", "index": i, "name": prop.prop_name})
				i += 1
		return props

class TFMetadataSetting(TFMetadataGroup):
	prop_names = []
	hide_custom = bpy.props.BoolProperty(default = True)

	@classmethod
	def register(cls):
		bpy.types.Object.tf_metadata = bpy.props.PointerProperty(type=cls)

	@classmethod
	def unregister(cls):
		del bpy.types.Object.tf_metadata

# Availability
class TFAvailabilitySetting(TFMetadataGroup):
	prop_names = ["yearFrom", "yearTo"]

	def get_yearFrom(self):
		if "yearFrom" in self:
			return self["yearFrom"]
		elif "yearFrom" in self.id_data:
			self["yearFrom"] = self.id_data["yearFrom"]
			return self.id_data["yearFrom"]
		return 0

	def set_yearFrom(self, value):
		self["yearFrom"] = value

	def get_yearTo(self):
		if "yearTo" in self:
			return self["yearTo"]
		elif "yearTo" in self.id_data:
			self["yearTo"] = self.id_data["yearTo"]
			return self.id_data["yearTo"]
		return 0

	def set_yearTo(self, value):
		self["yearTo"] = value

	yearFrom = bpy.props.IntProperty(default=0, get=get_yearFrom, set=set_yearFrom)
	yearTo = bpy.props.IntProperty(default=0, get=get_yearTo, set=set_yearTo)

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_availability", bpy.props.PointerProperty(type=cls, name="availability"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_availability")

# Cost
class TFCostSetting(TFMetadataGroup):
	prop_names = ["price"]
	def get_price(self):
		if "price" in self:
			return self["price"]
		elif "price" in self.id_data:
			self["price"] = self.id_data["price"]
			return self.id_data["price"]
		return 0

	def set_price(self, value):
		self["price"] = value

	price = bpy.props.IntProperty(get=get_price, set=set_price)

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_cost", bpy.props.PointerProperty(type=cls, name="cost"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_cost")

# Description
class TFDescriptionSetting(TFMetadataGroup):
	prop_names = [
		{"prop_name": "description_name", "name": "name"},
		"description",
		"smallIcon",
		"icon20",
		"icon",
	]

	def get_name(self):
		if "description_name" in self:
			return self["description_name"]
		elif "description_name" in self.id_data:
			self["description_name"] = self.id_data["description_name"]
			return self.id_data["description_name"]
		return ""

	def set_name(self, value):
		self["description_name"] = value

	def get_description(self):
		if "description" in self:
			return self["description"]
		elif "description" in self.id_data:
			self["description"] = self.id_data["description"]
			return self.id_data["description"]
		return ""

	def set_description(self, value):
		self["description"] = value

	description_name = bpy.props.StringProperty(get=get_name, set=set_name)
	description = bpy.props.StringProperty(get=get_description, set=set_description)
	smallIcon = bpy.props.StringProperty()
	icon20 = bpy.props.StringProperty()
	icon = bpy.props.StringProperty()

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_description", bpy.props.PointerProperty(type=cls, name="description"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_description")

# Maintenance
class TFMaintenanceSetting(TFMetadataGroup):
	prop_names = ["lifespan", "runningCosts"]
	def get_lifespan(self):
		if "lifespan" in self:
			return self["lifespan"]
		elif "lifespan" in self.id_data:
			self["lifespan"] = self.id_data["lifespan"]
			return self.id_data["lifespan"]
		return 0

	def set_lifespan(self, value):
		self["lifespan"] = value

	def get_runningCosts(self):
		if "runningCosts" in self:
			return self["runningCosts"]
		elif "runningCosts" in self.id_data:
			self["runningCosts"] = self.id_data["runningCosts"]
			return self.id_data["runningCosts"]
		return 0

	def set_runningCosts(self, value):
		self["runningCosts"] = value

	lifespan = bpy.props.IntProperty(min=0, get=get_lifespan, set=set_lifespan)
	runningCosts = bpy.props.IntProperty(get=get_runningCosts, set=set_runningCosts)

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_maintenance", bpy.props.PointerProperty(type=cls, name="maintenance"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_maintenance")

# ParticleSystem
class TFParticleSystemSetting(TFMetadataGroup):
	prop_names = []
	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_particleSystem", bpy.props.PointerProperty(type=cls, name="particleSystem"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_particleSystem")

# Particle Emitters
class TFParticleEmitterSetting(TFMetadataGroup):
	prop_names = [
		"position",
		"velocity",
		"color",
		"frequency",
		"lifeTime",
		"size01",
		"child"
	]

	def get_particleVar(self, newkey, oldkey, default):
		if newkey in self:
			return self[newkey]
		elif oldkey in self.id_data:
				self[newkey] = self.id_data[oldkey]
				return self.id_data[oldkey]
		return default

	def get_position(self):
		return self.get_particleVar("position", "particlePosition", Vector())

	def set_position(self, value):
		self["position"] = value

	def get_velocity(self):
		return self.get_particleVar("velocity", "particleVelocity", Vector())

	def set_velocity(self, value):
		self["velocity"] = value

	def get_color(self):
		return self.get_particleVar("color", "particleColor", Color())

	def set_color(self, value):
		self["color"] = value

	def get_frequency(self):
		return self.get_particleVar("frequency", "particleFrequency", 0)

	def set_frequency(self, value):
		self["frequency"] = value

	def get_lifeTime(self):
		return self.get_particleVar("lifeTime", "particleLifeTime", 0)

	def set_lifeTime(self, value):
		self["lifeTime"] = value

	def get_size01(self):
		return self.get_particleVar("size01", "particleSize01", Vector((0.0, 0.0)))

	def set_size01(self, value):
		self["size01"] = value

	def get_child(self):
		return self.get_particleVar("child", "particleChild", 1)

	def set_child(self, value):
		self["child"] = value

	position = bpy.props.FloatVectorProperty(size=3, default=[0.0, 0.0, 0.0], subtype='XYZ', get=get_position, set=set_position)
	velocity = bpy.props.FloatVectorProperty(size=3, default=[0.0, 0.0, 0.0], subtype='XYZ', get=get_velocity, set=set_velocity)
	color = bpy.props.FloatVectorProperty(size=3, default=[0.0, 0.0, 0.0], min=0.0, max=1.0, subtype='COLOR', get=get_color, set=set_color)
	frequency = bpy.props.IntProperty(default=0, get=get_frequency, set=set_frequency)
	lifeTime = bpy.props.FloatProperty(default=0, get=get_lifeTime, set=set_lifeTime, step=100, precision=1)
	size01 = bpy.props.FloatVectorProperty(size=2, default=[0.0, 0.0], get=get_size01, set=set_size01)
	child = bpy.props.IntProperty(default=1, get=get_child, set=set_child)
	emitterNum = bpy.props.IntProperty(min = 0)
	indexProp = "emitterNum"

	@classmethod
	def register(cls):
		TFParticleSystemSetting.add_prop("emitters", bpy.props.CollectionProperty(type = cls))
		TFParticleSystemSetting.add_prop("emitters_activeidx", bpy.props.IntProperty(min = 0), False)
	@classmethod
	def unregister(cls):
		TFParticleSystemSetting.remove_prop("emitters")
		TFParticleSystemSetting.remove_prop("emitters_activeidx")

class TFParticleEmitterAddOp(bpy.types.Operator):
	bl_label = "Add ParticleEmitter property"
	bl_idname = "collection.tfemitter_add"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_particleSystem.emitters
		item = collection.add()
		item.emitterNum = len(collection) - 1
		return {'FINISHED'}

class TFParticleEmitterRemoveOp(bpy.types.Operator):
	bl_label = "Remove ParticleEmitter property"
	bl_idname = "collection.tfemitter_remove"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_particleSystem.emitters
		index = obj.tf_metadata.tf_particleSystem.emitters_activeidx
		collection.remove(index)
		if index > 0:
			obj.tf_metadata.tf_particleSystem.emitters_activeidx = index-1
		return {'FINISHED'}

class TFParticleEmittersItemUIList(bpy.types.UIList):
	def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
		col = layout.box().column()
		col.row().prop(item, "position", "Positon")
		col.row().prop(item, "velocity", "Velocity")
		col.row().prop(item, "color", "Color")
		col.prop(item, "frequency", "Frequency")
		col.prop(item, "lifeTime", "Lifetime")
		col.row().prop(item, "size01", "Size")
		col.prop(item, "child", "Child")
		if not data.id_data.tf_metadata.hide_custom:
			draw_customMetadata(col, item, "model_metadata_particle_system_particle_emitter")

# RailVehicle
class TFRailVehicleSetting(TFMetadataGroup):
	prop_names = []
	def get_props(self):
		props = self.prop_names.copy()
		carrier = self.id_data.tf_metadata.tf_transportVehicle.carrier
		if carrier in _carrier_info:
			if "keys" in _carrier_info[carrier]:
				props = []
				for key in _carrier_info[carrier]["keys"]:
					props.append(key)

		if hasattr(self, "customMetadata"):
			i = 0
			for prop in self.customMetadata:
				props.append({"prop_name": "customMetadata", "index": i, "name": prop.prop_name})
				i += 1

		return props

	def get_topSpeed(self):
		if "topSpeed" in self:
			return self["topSpeed"]
		elif "topSpeed" in self.id_data:
			self["topSpeed"] = self.id_data["topSpeed"]
			return self.id_data["topSpeed"]
		return 0.0

	def set_topSpeed(self, value):
		self["topSpeed"] = value

	def get_weight(self):
		if "weight" in self:
			return self["weight"]
		elif "weight" in self.id_data:
			self["weight"] = self.id_data["weight"]
			return self.id_data["weight"]
		return 0.0

	def set_weight(self, value):
		self["weight"] = value

	def get_power(self):
		if "power" in self:
			return self["power"]
		elif len(self.engines) > 0:
			self["power"] = self.engines[0].enginePower
			return self["power"]
		return 0.0

	def set_power(self, value):
		self["power"] = value

	def get_effort(self):
		if "tractiveEffort" in self:
			return self["tractiveEffort"]
		elif len(self.engines) > 0:
			self["tractiveEffort"] = self.engines[0].engineEffort
			return self["tractiveEffort"]
		return 0.0

	def set_effort(self, value):
		self["tractiveEffort"] = value

	topSpeed = bpy.props.FloatProperty(min=0.0, get=get_topSpeed, set=set_topSpeed, step=100, precision=1)
	weight = bpy.props.FloatProperty(min=0.0, get=get_weight, set=set_weight, step=100, precision=1)
	power = bpy.props.FloatProperty(min=0.0, get=get_power, set=set_power, step=100)
	tractiveEffort = bpy.props.FloatProperty(min=0.0, get=get_effort, set=set_effort, step=100)
	maxRpm = bpy.props.FloatProperty(min=0.0, step=100)
	area = bpy.props.FloatProperty(min=0.0, step=100)
	maxPayload = bpy.props.FloatProperty(min=0.0)
	maxTakeOffWeight = bpy.props.FloatProperty(min=0.0, step=100)
	timeToFullThrust = bpy.props.FloatProperty(min=0.0, step=100)

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_railVehicle", bpy.props.PointerProperty(type=cls, name="vehicle"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_railVehicle")

# Engines
class TFEngineSetting(TFMetadataGroup):
	prop_names = [
		{"prop_name": "engineType", "name": "type"},
		{"prop_name": "enginePower", "name": "power"},
		{"prop_name": "engineEffort", "name": "tractiveEffort"},
	]

	def get_engine_list(self, key):
		engines = []
		for child in self.id_data.children:
			if key in child:
				engines.append(child[key])
		return engines

	def get_engType(self):
		if "engineType" in self:
			return self["engineType"]
		else:
			engines = self.get_engine_list("engineType")
			if self.engineNum < len(engines):
				self["engineType"] = engines[self.engineNum]
				return engines[self.engineNum]
		return 0

	def set_engType(self, value):
		self["engineType"] = value

	def get_engPower(self):
		if "enginePower" in self:
			return self["enginePower"]
		else:
			engines = self.get_engine_list("enginePower")
			if self.engineNum < len(engines):
				self["enginePower"] = engines[self.engineNum]
				return engines[self.engineNum]
		return 0

	def set_engPower(self, value):
		self["enginePower"] = value

	def get_engEffort(self):
		if "engineEffort" in self:
			return self["engineEffort"]
		else:
			engines = self.get_engine_list("engineTractiveEffort")
			if self.engineNum < len(engines):
				self["engineEffort"] = engines[self.engineNum]
				return engines[self.engineNum]
		return 0

	def set_engEffort(self, value):
		self["engineEffort"] = value

	engineType = bpy.props.EnumProperty(
		items=[(_engine_list[i][1], _engine_list[i][0], _engine_list[i][2], i) for i in range(0, len(_engine_list))],
		get = get_engType,
		set = set_engType
	)
	enginePower = bpy.props.FloatProperty(min=0.0, get=get_engPower, set=set_engPower, step=100)
	engineEffort = bpy.props.FloatProperty(min=0.0, get=get_engEffort, set=set_engEffort, step=100, precision=1)
	engineNum = bpy.props.IntProperty(min = 0)
	indexProp = "engineNum"

	@classmethod
	def register(cls):
		TFRailVehicleSetting.add_prop("engines", bpy.props.CollectionProperty(type = cls), False)
		TFRailVehicleSetting.add_prop("engines_activeidx", bpy.props.IntProperty(min = 0), False)

	@classmethod
	def unregister(cls):
		TFRailVehicleSetting.remove_prop("engines")
		TFRailVehicleSetting.remove_prop("engines_activeidx")

class TFEngineAddOp(bpy.types.Operator):
	bl_label = "Add engine property"
	bl_idname = "collection.tfengine_add"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_railVehicle.engines
		item = collection.add()
		item.engineNum = len(collection) - 1
		return {'FINISHED'}

class TFEngineRemoveOp(bpy.types.Operator):
	bl_label = "Remove engine property"
	bl_idname = "collection.tfengine_remove"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_railVehicle.engines
		index = obj.tf_metadata.tf_railVehicle.engines_activeidx
		collection.remove(index)
		if index > 0:
			obj.tf_metadata.tf_railVehicle.engines_activeidx = index-1
		return {'FINISHED'}

class TFEngineSettingsItemUIList(bpy.types.UIList):
	def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
		layout.prop(item, "engineType", "Type")
		layout.prop(item, "enginePower", "Power")
		layout.prop(item, "engineEffort", "Tractive Effort")
		if not data.id_data.tf_metadata.hide_custom:
			draw_customMetadata(layout, item, "model_metadata_vehicle_engine")

# SoundSet
class TFSoundSetSetting(TFMetadataGroup):
	prop_names = [
		{"prop_name": "soundSet_name", "name": "name"},
		"horn"
	]
	def get_name(self):
		if "soundSet_name" in self:
			return self["soundSet_name"]
		elif "soundSet_name" in self.id_data:
			self["soundSet_name"] = self.id_data["soundSet_name"]
			return self.id_data["soundSet_name"]
		return ""

	def set_name(self, value):
		self["soundSet_name"] = value

	def get_horn(self):
		if "horn" in self:
			return self["horn"]
		elif "soundSet_horn" in self.id_data:
			self["horn"] = self.id_data["soundSet_horn"]
			return self.id_data["soundSet_horn"]
		return ""

	def set_horn(self, value):
		self["horn"] = value

	soundSet_name = bpy.props.StringProperty(default='', get=get_name, set=set_name)
	horn = bpy.props.StringProperty(default='', get=get_horn, set=set_horn)
	singleValue = "soundSet_name"

	@classmethod
	def register(cls):
		TFRailVehicleSetting.add_prop("soundSet", bpy.props.PointerProperty(type=cls), False)

	@classmethod
	def unregister(cls):
		TFRailVehicleSetting.remove_prop("soundSet")

# TransportVehicle
class TFTransportVehicleSetting(TFMetadataGroup):
	prop_names = ["carrier", "loadSpeed", "multipleUnitOnly"]
	def get_loadSpeed(self):
		if "loadSpeed" in self:
			return self["loadSpeed"]
		elif "loadSpeed" in self.id_data:
			self["loadSpeed"] = self.id_data["loadSpeed"]
			return self.id_data["loadSpeed"]
		return 1

	def set_loadSpeed(self, value):
		self["loadSpeed"] = value

	carrier = bpy.props.EnumProperty(items=[(_carrier_list[i][1], _carrier_list[i][0], _carrier_list[i][2], i) for i in range(0, len(_carrier_list))])
	loadSpeed = bpy.props.FloatProperty(min=0.0, default=1.0, step=50, precision=1, get=get_loadSpeed, set=set_loadSpeed)
	multipleUnitOnly = bpy.props.BoolProperty(default=False)

	@classmethod
	def register(cls):
		TFMetadataSetting.add_prop("tf_transportVehicle", bpy.props.PointerProperty(type=cls, name="transportVehicle"))

	@classmethod
	def unregister(cls):
		TFMetadataSetting.remove_prop("tf_transportVehicle")

# Capacities
class TFCapacitySetting(TFMetadataGroup):
	prop_names = [
		{"prop_name": "capacityType", "name": "type"},
		"capacity",
	]
	def get_capacity_list(self, key):
		capacities = []
		for child in self.id_data.children:
			if key in child:
				capacities.append(child[key])
		return capacities

	def get_capacityType(self):
		capacityID = ""
		if "capacityType" in self:
			capacityID = self["capacityType"]
		else:
			capacities = self.get_capacity_list("capacityType")
			if self.capacityNum < len(capacities):
				self["capacityType"] = _cargo_type_list[capacities[self.capacityNum]]
				capacityID = _cargo_type_list[capacities[self.capacityNum]]

		if capacityID in _cargo_type_list:
			return _cargo_type_list.index(capacityID)
		return 0

	def set_capacityType(self, value):
		self["capacityType"] = _cargo_type_list[value]

	def get_capacity(self):
		if "capacity" in self:
			return self["capacity"]
		else:
			capacities = self.get_capacity_list("capacity")
			if self.capacityNum < len(capacities):
				self["capacity"] = capacities[self.capacityNum]
				return capacities[self.capacityNum]
		return 0

	def set_capacity(self, value):
		self["capacity"] = value

	capacityType = bpy.props.EnumProperty(
		items=[(_cargo_type_list_enum[i][1], _cargo_type_list_enum[i][0], _cargo_type_list_enum[i][2], i) for i in range(0, len(_cargo_type_list_enum))],
		get = get_capacityType,
		set = set_capacityType
	)
	capacity = bpy.props.IntProperty(min=0, get=get_capacity, set=set_capacity)
	capacityNum = bpy.props.IntProperty(min = 0)
	indexProp = "capacityNum"

	@classmethod
	def register(cls):
		TFTransportVehicleSetting.add_prop("capacities", bpy.props.CollectionProperty(type = cls))
		TFTransportVehicleSetting.add_prop("capacities_activeidx", bpy.props.IntProperty(min = 0), False)

	@classmethod
	def unregister(cls):
		TFTransportVehicleSetting.remove_prop("capacities")
		TFTransportVehicleSetting.remove_prop("capacities_activeidx")

class TFCapacityAddOp(bpy.types.Operator):
	bl_label = "Add capacity property"
	bl_idname = "collection.tfcapacity_add"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_transportVehicle.capacities
		item = collection.add()
		item.capacityNum = len(collection) - 1
		return {'FINISHED'}

class TFCapacityRemoveOp(bpy.types.Operator):
	bl_label = "Remove capacity property"
	bl_idname = "collection.tfcapacity_remove"

	def invoke(self, context, event):
		obj = context.object
		collection = obj.tf_metadata.tf_transportVehicle.capacities
		index = obj.tf_metadata.tf_transportVehicle.capacities_activeidx
		collection.remove(index)
		if index > 0:
			obj.tf_metadata.tf_transportVehicle.capacities_activeidx = index-1
		return {'FINISHED'}

class TFCapacitySettingsItemUIList(bpy.types.UIList):
	def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
		layout.prop(item, "capacityType", "Type")
		layout.prop(item, "capacity", "Capacity")
		if not data.id_data.tf_metadata.hide_custom:
			draw_customMetadata(layout, item, "model_metadata_transport_vehicle_capacity")

# custom metadata
class TFCustomMetadataSetting(TFPropertyGroup):
	prop_name = bpy.props.StringProperty()
	prop_type = bpy.props.EnumProperty(items=_tfproperty_types, default="prop_string")
	prop_intValue = bpy.props.IntProperty()
	prop_floatValue = bpy.props.FloatProperty()
	prop_boolValue = bpy.props.BoolProperty()
	prop_stringValue = bpy.props.StringProperty()
	prop_listValue_activeidx = bpy.props.IntProperty(min = 0)

	class_list = [
		TFMetadataSetting,
		TFAvailabilitySetting,
		TFCostSetting,
		TFDescriptionSetting,
		TFMaintenanceSetting,
		TFParticleSystemSetting,
		TFParticleEmitterSetting,
		TFRailVehicleSetting,
		TFEngineSetting,
		TFTransportVehicleSetting,
		TFCapacitySetting,
		TFSoundSetSetting
	]

	@classmethod
	def register(cls):
		for class_type in cls.class_list:
			class_type.add_prop("customMetadata", bpy.props.CollectionProperty(type=cls), False)
			class_type.add_prop("customMetadata_activeidx", bpy.props.IntProperty(min = 0), False)
		TFCustomMetadataSetting.add_prop("prop_listValue", bpy.props.CollectionProperty(type=TFCustomMetadataSetting), False)

	@classmethod
	def unregister(cls):
		for i in range(len(cls.class_list)-1, -1, -1):
			cls.class_list[i].remove_prop("customMetadata")
			cls.class_list[i].remove_prop("customMetadata_activeidx")
		TFCustomMetadataSetting.remove_prop("prop_listValue")

	@classmethod
	def getProptype(cls, proptype):
		if proptype in {"<class 'int'>", "<class 'long'>"}:
			return "prop_int"
		elif proptype == "<class 'float'>":
			return "prop_float"
		elif proptype == "<class 'str'>":
			return "prop_string"
		elif proptype == "<class 'bool'>":
			return "prop_bool"
		elif proptype in ("<class 'dict'>", "<class 'list'>"):
			return "prop_table"
		else:
			return "empty"

	def to_table(self):
		result = []
		i = 0
		for item in self.prop_listValue:
			if item.prop_type != "empty":
				if isinstance(result, list):
					if item.prop_name == "":
						result.append(item.getValue())
					else:
						result = dict(zip(range(len(result)), result))
						i = len(result)
						result[item.prop_name] = item.getValue()
				else:
					if item.prop_name == "":
						result[i] = item.getValue()
						i += 1
					else:
						result[item.prop_name] = item.getValue()
		return result


	def setValue(self, value):
		ptype = self.prop_type
		if ptype == "prop_string":
			self.prop_stringValue = value
		elif ptype == "prop_int":
			self.prop_intValue = value
		elif ptype == "prop_float":
			self.prop_floatValue = value
		elif ptype == "prop_bool":
			self.prop_boolValue = value
		elif ptype == "prop_table":
			while len(self.prop_listValue) > 0:
				self.prop_listValue.remove(0)

			if len(value) > 0:
				if isinstance(value, dict):
					for key in value.keys():
						item = self.prop_listValue.add()
						item.prop_name = str(key)
						item.prop_type = TFCustomMetadataSetting.getProptype(str(type(value[key])))
						item.setValue(value[key])

						item_value = item.getValue()
						if item_value == None:
							self.prop_listValue.remove(len(self.prop_listValue)-1)
						elif isinstance(item_value, dict) or isinstance(item_value, list):
							if len(item_value) == 0:
								self.prop_listValue.remove(len(self.prop_listValue)-1)
				else:
					for var in value:
						item = self.prop_listValue.add()
						item.prop_name = ""
						item.prop_type = TFCustomMetadataSetting.getProptype(str(type(var)))
						item.setValue(var)

						item_value = item.getValue()
						if item_value == None:
							self.prop_listValue.remove(len(self.prop_listValue)-1)
						elif isinstance(item_value, dict) or isinstance(item_value, list):
							if len(item_value) == 0:
								self.prop_listValue.remove(len(self.prop_listValue)-1)
			else:
				self.prop_type = "empty"

	def getValue(self):
		ptype = self.prop_type
		if ptype == "prop_string":
			return self.prop_stringValue
		elif ptype == "prop_int":
			return self.prop_intValue
		elif ptype == "prop_float":
			return self.prop_floatValue
		elif ptype == "prop_bool":
			return self.prop_boolValue
		elif ptype == "prop_table":
			return self.to_table()
		elif ptype == "empty":
			return None

		raise ValueError('invalid type : "%s"' % ptype)

	def draw(self, layout, attr=""):
		ptype = self.prop_type
		if ptype == "prop_table":
			col = layout.column()
			row = col.row()
			row.prop(self, "prop_name", "Name")
			row.prop(self, "prop_type", "")
			row = col.row()
			listId = self.path_from_id()
			listId = listId.replace(".prop_listValue", "")
			listId = listId.replace(".", "_")
			listId = listId.replace("[", "_")
			listId = listId.replace("]", "")
			row.template_list(
				listtype_name="TFCustomMetadataSettingsItemUIList",
				dataptr=self,
				propname="prop_listValue",
				active_dataptr=self,
				active_propname="prop_listValue_activeidx",
				rows = 1,
				maxrows = 3,
				list_id = listId+"_custom_metadata",
			)
			col = row.column(align=True)
			col.operator("collection.tfcustommetadata_add", icon="ZOOMIN", text="").attr = self.path_from_id()+".prop_listValue"
			col.operator("collection.tfcustommetadata_remove", icon="ZOOMOUT", text="").attr = self.path_from_id()+".prop_listValue"
		else:
			layout.prop(self, "prop_name", "Name")
			layout.prop(self, "prop_type", "")
			if ptype == "prop_string":
				layout.prop(self, "prop_stringValue", "Value")
			elif ptype == "prop_int":
				layout.prop(self, "prop_intValue", "Value")
			elif ptype == "prop_float":
				layout.prop(self, "prop_floatValue", "Value")
			elif ptype == "prop_bool":
				layout.prop(self, "prop_boolValue", "Value")

class TFCustomMetadataAddOp(bpy.types.Operator):
	bl_label = "Add custom metadata property"
	bl_idname = "collection.tfcustommetadata_add"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		obj = context.object
		attrs = self.attr.split(".")
		for attribute in attrs:
			indices = attribute.split("[")[1:]
			attribute = attribute.split("[")[0]
			if hasattr(obj, attribute):
				obj = getattr(obj, attribute)
			else:
				break
			for index in indices:
				index = int(index[:-1])
				obj = obj[index]
		else:
			collection = obj
			item = collection.add()
			item.attr = self.attr + "[%s]" % str(len(collection)-1)
		return {'FINISHED'}

class TFCustomMetadataRemoveOp(bpy.types.Operator):
	bl_label = "Remove custom metadata property"
	bl_idname = "collection.tfcustommetadata_remove"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		obj = context.object
		attrs = self.attr.split(".")
		for i in range(len(attrs)-1):
			indices = attrs[i].split("[")[1:]
			attribute = attrs[i].split("[")[0]
			if hasattr(obj, attribute):
				obj = getattr(obj, attribute)
			else:
				break
			for index in indices:
				index = int(index[:-1])
				obj = obj[index]
		else:
			index = 0
			if hasattr(obj, attrs[i+1] + "_activeidx"):
				index = getattr(obj, attrs[i+1] + "_activeidx")
			if hasattr(obj, attrs[i+1]):
				collection = getattr(obj, attrs[i+1])
				collection.remove(index)
		return {'FINISHED'}

class TFCustomMetadataSettingsItemUIList(bpy.types.UIList):
	def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
		item.draw(layout)

class TFMeshMetadataSetting(TFPropertyGroup):
	def _get_axles(self):
		if "axles" in self:
			return self["axles"]
		elif "isAxle" in self.id_data.tf_mesh:
			self["axles"] = self.id_data.tf_mesh["isAxle"]
			del self.id_data.tf_mesh["isAxle"]
			return self["axles"]
		elif "isAxle" in self.id_data:
			self["axles"] = self.id_data["isAxle"]
			del self.id_data["isAxle"]
			return self["axles"]
		return False

	def _set_axles(self, value):
		self["axles"] = value

	def _get_wheels(self):
		if "wheels" in self:
			return self["wheels"]
		elif "isWheel" in self.id_data.tf_mesh:
			self["wheels"] = self.id_data.tf_mesh["isWheel"]
			del self.id_data.tf_mesh["isWheel"]
			return self["wheels"]
		elif "isWheel" in self.id_data:
			self["wheels"] = self.id_data["isWheel"]
			del self.id_data["isWheel"]
			return self["wheels"]
		return False

	def _set_wheels(self, value):
		self["wheels"] = value

	@classmethod
	def register(cls):
		for key in _mesh_config_keys:
			if hasattr(cls, "_get_"+key[0]):
				setattr(cls, key[0], bpy.props.BoolProperty(default=False, get = getattr(cls, "_get_"+key[0]), set = getattr(cls, "_set_"+key[0])))
			else:
				setattr(cls, key[0], bpy.props.BoolProperty(default=False))

		TFMeshSetting.metadata = bpy.props.PointerProperty(type = cls)

	@classmethod
	def unregister(cls):
		del TFMeshSetting.metadata
		for key in _mesh_config_keys:
			delattr(cls, key)

class TFLodMetadataSetting(TFPropertyGroup):
	@classmethod
	def register(cls):
		for key in _lod_config_keys:
			setattr(cls, key[0], bpy.props.PointerProperty(type = key[2]))
		TFLodSetting.metadata = bpy.props.PointerProperty(type = cls)

	@classmethod
	def unregister(cls):
		del TFLodSetting.metadata
		for key in _lod_config_keys:
			delattr(cls, key)

class TFMetadataImportExport():
	## Internal functions
	# Panel drawing functions
	def __draw_particleSystem(self, layout, mydata):
		boxPartSys = TF_base_functions.draw_hideable_box(layout, "ParticleSystem:", "model_metadata_particle_system", False)
		if boxPartSys:
			row = boxPartSys.row()
			row.template_list(
				listtype_name="TFParticleEmittersItemUIList",
				dataptr=mydata,
				propname="emitters",
				active_dataptr=mydata,
				active_propname="emitters_activeidx",
				rows = 1,
				maxrows = 2
			)
			col = row.column(align=True)
			col.operator("collection.tfemitter_add", icon="ZOOMIN", text="")
			col.operator("collection.tfemitter_remove", icon="ZOOMOUT", text="")
			if not mydata.id_data.tf_metadata.hide_custom:
				draw_customMetadata(boxPartSys, mydata, "model_metadata_particle_system")
			layout.separator()

	def __draw_railVehicle(self, layout, mydata, carrier):
		boxRailV = TF_base_functions.draw_hideable_box(layout, "Vehicle:", "model_metadata_vehicle", False)
		if boxRailV:
			if carrier in _carrier_info and "keys" in _carrier_info[carrier]:
				for key in _carrier_info[carrier]["keys"]:
					text = None
					if isinstance(key, dict):
						prop_name = key["prop_name"]
						if "text" in key:
							text = key["text"]
					else:
						prop_name = key

					if prop_name == "engines":
						self.__draw_engine(boxRailV, mydata)
					elif prop_name == "soundSet":
						boxSoundSet = TF_base_functions.draw_hideable_box(boxRailV, "SoundSet:", "model_metadata_vehicle_sound_set", False)
						if boxSoundSet:
							boxSoundSet.prop(mydata.soundSet, "horn", text="Horn")
							boxSoundSet.prop(mydata.soundSet, "soundSet_name", text="Name")
							if not mydata.id_data.tf_metadata.hide_custom:
								draw_customMetadata(boxSoundSet, mydata.soundSet, "model_metadata_vehicle_sound_set")
					else:
						if text != None:
							boxRailV.prop(mydata, prop_name, text=text)
						else:
							boxRailV.prop(mydata, prop_name)

			if not mydata.id_data.tf_metadata.hide_custom:
				draw_customMetadata(boxRailV, mydata, "model_metadata_vehicle")
			layout.separator()

	def __draw_engine(self, layout, mydata):
		boxEng = TF_base_functions.draw_hideable_box(layout, "Engines:", "model_metadata_vehicle_engines", False)
		if boxEng:
			row = boxEng.row()
			row.template_list(
				listtype_name="TFEngineSettingsItemUIList",
				dataptr=mydata,
				propname="engines",
				active_dataptr=mydata,
				active_propname="engines_activeidx",
				rows = 1,
				maxrows = 2
			)
			col = row.column(align=True)
			col.operator("collection.tfengine_add", icon="ZOOMIN", text="")
			col.operator("collection.tfengine_remove", icon="ZOOMOUT", text="")

	def __draw_transportVehicle(self, layout, mydata):
		boxTranspV = TF_base_functions.draw_hideable_box(layout, "TransportVehicle:", "model_metadata_transport_vehicle", False)
		if boxTranspV:
			self.__draw_capacity(boxTranspV, mydata)
			boxTranspV.prop(mydata, "carrier", text="Carrier")
			if len(mydata.capacities) > 0:
				boxTranspV.prop(mydata, "loadSpeed", text="Load Speed")
			boxTranspV.prop(mydata, "multipleUnitOnly", text="Multiple Unit only")
			if not mydata.id_data.tf_metadata.hide_custom:
				draw_customMetadata(boxTranspV, mydata, "model_metadata_transport_vehicle")
			if not mydata.id_data.tf_metadata.hide_custom:
				layout.separator()

	def __draw_capacity(self, layout, mydata):
		boxCapa = TF_base_functions.draw_hideable_box(layout, "Capacities:", "model_metadata_transport_vehicle_capacities", False)
		if boxCapa:
			row = boxCapa.row()
			row.template_list(
				listtype_name="TFCapacitySettingsItemUIList",
				dataptr=mydata,
				propname="capacities",
				active_dataptr=mydata,
				active_propname="capacities_activeidx",
				rows = 1,
				maxrows = 2
			)
			col = row.column(align=True)
			col.operator("collection.tfcapacity_add", icon="ZOOMIN", text="")
			col.operator("collection.tfcapacity_remove", icon="ZOOMOUT", text="")

	def __draw_lod_metadata(self, layout, mydata):
		mydata = mydata.tf_lod.metadata
		boxConfig = TF_base_functions.draw_hideable_box(layout, "Configs:", "lod_metadata_configs", False)
		if boxConfig:
			for key in _lod_config_keys:
				boxKey = TF_base_functions.draw_hideable_box(boxConfig, key[1]+":", "lod_metadata_configs_"+key[0])
				if boxKey:
					key[3](boxKey, getattr(mydata, key[0]), key[0])
					boxConfig.separator()
			layout.separator()

	def __draw_model_metadata(self, layout, mydata):
		mydata = mydata.tf_metadata
		boxMeta = TF_base_functions.draw_hideable_box(layout, "Metadata:", "model_metadata", False)
		if boxMeta:
			boxMeta.prop(mydata, "hide_custom", text="Hide custom entries")
			boxAvail = TF_base_functions.draw_hideable_box(boxMeta, "Availability:", "model_metadata_availability", False)
			if boxAvail:
				boxAvail.prop(mydata.tf_availability, "yearFrom", text="Start Year")
				boxAvail.prop(mydata.tf_availability, "yearTo", text="End Year")
				if not mydata.hide_custom:
					draw_customMetadata(boxAvail, mydata.tf_availability, "model_metadata_availability")
				boxMeta.separator()

			boxCost = TF_base_functions.draw_hideable_box(boxMeta, "Cost:", "model_metadata_cost", False)
			if boxCost:
				boxCost.prop(mydata.tf_cost, "price", text="Price")
				if not mydata.hide_custom:
					draw_customMetadata(boxCost, mydata.tf_cost, "model_metadata_cost")
				boxMeta.separator()

			boxDesc = TF_base_functions.draw_hideable_box(boxMeta, "Description:", "model_metadata_description", False)
			if boxDesc:
				boxDesc.prop(mydata.tf_description, "description_name", text="Name")
				boxDesc.prop(mydata.tf_description, "description", text="Description")
				boxDesc.prop(mydata.tf_description, "smallIcon", text="UI Icon")
				if not mydata.hide_custom:
					draw_customMetadata(boxDesc, mydata.tf_description, "model_metadata_description")
				boxMeta.separator()

			boxMaint = TF_base_functions.draw_hideable_box(boxMeta, "Maintenance:", "model_metadata_maintenance", False)
			if boxMaint:
				boxMaint.prop(mydata.tf_maintenance, "lifespan", text="Lifespan")
				boxMaint.prop(mydata.tf_maintenance, "runningCosts", text="Running Costs")
				if not mydata.hide_custom:
					draw_customMetadata(boxMaint, mydata.tf_maintenance, "model_metadata_maintenance")
				boxMeta.separator()

			self.__draw_particleSystem(boxMeta, mydata.tf_particleSystem)

			self.__draw_railVehicle(boxMeta, mydata.tf_railVehicle, mydata.tf_transportVehicle.carrier)

			self.__draw_transportVehicle(boxMeta, mydata.tf_transportVehicle)

			if not mydata.hide_custom:
				draw_customMetadata(boxMeta, mydata, "model_metadata")

	# data import helper functions
	def __create_prop(self, obj, name, proptype):
		item = obj.customMetadata.add()
		item.prop_name = name
		item.prop_type = TFCustomMetadataSetting.getProptype(proptype)
		return item

	def __remove_prop(self, obj, index=-1):
		if index == -1:
			index = len(obj.customMetadata)-1
		obj.customMetadata.remove(index)

	def __asign_to_prop(self, data, prop, attr):
		vartype = str(type(data))
		if prop:
			if vartype in {"<class 'int'>", "<class 'long'>", "<class 'float'>", "<class 'str'>", "<class 'bool'>"}:
				# "normal" values
				if hasattr(prop, attr):
					new_prop = getattr(prop, attr)
					if get_basetype(vartype) == get_basetype(str(type(new_prop))):
						setattr(prop, attr, data)
					elif hasattr(new_prop, "singleValue") and hasattr(new_prop, new_prop.singleValue):
						if get_basetype(vartype) == get_basetype(str(type(getattr(new_prop, new_prop.singleValue)))):
							setattr(new_prop, new_prop.singleValue, data)

			elif vartype == "<class 'dict'>":
				# dictionary
				if attr != "":
					prop = getattr(prop, attr)

				prop_list = {}
				for key in prop.get_props():
					if isinstance(key, dict):
						if "name" in key:
							prop_list[key["name"]] = {"prop_name": key["prop_name"]}
						else:
							prop_list[key["prop_name"]] = {"prop_name": key["prop_name"]}
					else:
						prop_list[key] = {"prop_name": key}

				for key in sorted(data.keys()):
					if key not in prop_list:
						# no predefined prop available, use custom metadata
						item = self.__create_prop(prop, key, str(type(data[key])))
						item.setValue(data[key])
						item_value = item.getValue()

						# remove empty custom props
						if item_value == None:
							self.__remove_prop(prop)
						elif isinstance(item_value, dict) or isinstance(item_value, list):
							if len(item_value) == 0:
								self.__remove_prop(prop)

					else:
						#use predefined prop
						self.__asign_to_prop(data[key], prop, prop_list[key]["prop_name"])

			elif vartype == "<class 'list'>":
				# Array/List
				if attr != "":
					new_prop = getattr(prop, attr)
				else:
					new_prop = prop

				if str(type(new_prop)) == "<class 'bpy_prop_collection_idprop'>":
					for var in data:
						modelprop = new_prop.add()
						if hasattr(modelprop, "indexProp"):
							setattr(modelprop, modelprop.indexProp, len(new_prop)-1)

						new_attr = ""
						if hasattr(modelprop, "valueProp"):
							new_attr = modelprop.valueProp
						self.__asign_to_prop(var, modelprop, new_attr)

						# remove empty custom props
						if hasattr(modelprop, "getValue"):
							modelprop_value = modelprop.getValue()
							if modelprop_value == None:
								self.__remove_prop(prop)
							elif isinstance(modelprop_value, dict) or isinstance(modelprop_value, list):
								if len(modelprop_value) == 0:
									self.__remove_prop(prop)

				elif hasattr(prop, attr) and len(data) > 0:
					setattr(prop, attr, data)

	# get wheel and axle data from objects
	def __get_configs_from_group_TPF2(self, model_data, children, config):
		meshes = []
		if "meshes" in TF_base_functions.export_settings["filedata"]:
			meshes = TF_base_functions.export_settings["filedata"]["meshes"]
		groups = []
		if "groups" in TF_base_functions.export_settings["filedata"]:
			groups = TF_base_functions.export_settings["filedata"]["groups"]

		for child in children:
			if "mesh" in child: 
				# (base, filename) = TF_base_functions.split_path(child["mesh"])
				meshWithPrefix = child["mesh"]
				meshWithoutPrefix = meshWithPrefix
				if TF_base_functions.export_settings["prefix"] != "":
					base, discard, filename = child["mesh"].rpartition(TF_base_functions.export_settings["prefix"])
					meshWithoutPrefix = base + filename
				# print("Seeking %s in:" % meshWithoutPrefix)
				# for mesh in meshes:
					# print("\t" + mesh)
				
				if meshWithoutPrefix[:-4] in meshes:
					
					mesh = meshes[meshWithoutPrefix[:-4]]
					
					if "__metadata_config" in mesh["data"]:
						# if (TF_base_functions.user_prefs["debug"]):
							# print("\tFound __metadata_config in %s" % meshWithPrefix)
						if mesh["data"]["__metadata_config"] not in config:
							config[mesh["data"]["__metadata_config"]] = []
							# if (TF_base_functions.user_prefs["debug"]):
								# print("\tAdding config %s of %s to the list of configs which will be printed" % (mesh["data"]["__metadata_config"], meshWithPrefix))
						if meshWithPrefix not in config[mesh["data"]["__metadata_config"]]:
							config[mesh["data"]["__metadata_config"]].append(meshWithPrefix)
							# if (TF_base_functions.user_prefs["debug"]):
								# print("\tAdding %s as an element of config %s" % (meshWithPrefix, mesh["data"]["__metadata_config"]))
			
			if "children" in child:
				self.__get_configs_from_group_TPF2(model_data, child["children"], config)
					
	def __get_configs_from_group(self, model_data, children, config):
		meshes = []
		if "meshes" in TF_base_functions.export_settings["filedata"]:
			meshes = TF_base_functions.export_settings["filedata"]["meshes"]
		groups = []
		if "groups" in TF_base_functions.export_settings["filedata"]:
			groups = TF_base_functions.export_settings["filedata"]["groups"]

		for child in children:
			# THIS SECTION LIKELY WILL NOT WORK WITH "PREFIX" SET - see __get_configs_from_group_TPF2 for clues on how to get around that
		
			if child["type"] == "MESH" and child["id"][:-4] in meshes:
				mesh = meshes[child["id"][:-4]]
				if "__metadata_config" in mesh["data"]:
					if mesh["data"]["__metadata_config"] not in config:
						config[mesh["data"]["__metadata_config"]] = []
					if child["id"] not in config[mesh["data"]["__metadata_config"]]:
						config[mesh["data"]["__metadata_config"]].append(child["id"])
			elif child["type"] == "GROUP" and child["id"][:-4] in groups:
				group = groups[child["id"][:-4]]
				if "children" in group["data"]:
					self.__get_configs_from_group(model_data, group["data"]["children"], config)

	## External functions
	# UI functions
	def draw_lod_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport draw_lod_settings_panel")
		self.__draw_lod_metadata(data["layout"], data["mydata"])
	
	def draw_model_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport draw_model_settings_panel")
		self.__draw_model_metadata(data["layout"], data["mydata"])

	def draw_mesh_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport draw_mesh_settings_panel")

		boxConfig = TF_base_functions.draw_hideable_box(data["layout"], "Configs:", "mesh_metadata_configs", False)
		if boxConfig:
			for key in _mesh_config_keys:
				boxConfig.prop(data["mydata"].tf_mesh.metadata, key[0], text=key[1])
			data["layout"].separator()

	# import/export functions
	def save_mesh_data_once(self, data):
		for key in _mesh_config_keys:
			if getattr(data["meshObj"].data.tf_mesh.metadata, key[0]):
				data["meshData"]["__metadata_config"] = key[0]
				break

	def save_lod_data(self, data):
		lodData = data["lodData"]
		lod = data["lodObj"]
		
		if lod:
			lodData["__configs"] = {}
			for key in _lod_config_keys:
				if hasattr(lod.tf_lod.metadata, key[0]):
					value = key[5](getattr(lod.tf_lod.metadata, key[0]), data["children"])
					if value:
						lodData["__configs"][key[0]] = value

	def save_model_data(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport save_model")

		model_data = data["modelData"]
		if "metadata" in model_data:
			new_metadata = TF_base_functions.prop_to_value(data["modelObj"].tf_metadata)
			for key in new_metadata.keys():
				if key in model_data["metadata"] and isinstance(model_data["metadata"][key], dict) and isinstance(new_metadata[key], dict):
					model_data["metadata"][key].update(new_metadata[key])
				else:
					model_data["metadata"][key] = new_metadata[key]
		else:
			model_data["metadata"] = TF_base_functions.prop_to_value(data["modelObj"].tf_metadata)

		if "description" in model_data["metadata"]:
			if "name" in model_data["metadata"]["description"]:
				model_data["metadata"]["description"]["name"] = "_(\"" + model_data["metadata"]["description"]["name"] + "\")"
			if "description" in model_data["metadata"]["description"]:
				model_data["metadata"]["description"]["description"] = "_(\"" + model_data["metadata"]["description"]["description"] + "\")"

		carrier = "NONE"
		if "transportVehicle" in model_data["metadata"]:
			if "carrier" in model_data["metadata"]["transportVehicle"]:
				carrier = model_data["metadata"]["transportVehicle"]["carrier"]
		elif "car" in model_data["metadata"]:
			carrier = "ROAD"

		# get carrier specific information
		vehicletype = ""
		if carrier in _carrier_info:
			vehicletype = _carrier_info[carrier]["vehicletype"]
			
		# change soundset definition for planes and ships
		if "vehicle" in model_data["metadata"] and carrier in ("WATER", "AIR"):
			if "soundSet" in model_data["metadata"]["vehicle"]:
				model_data["metadata"]["soundConfig"] = {"soundSet": model_data["metadata"]["vehicle"]["soundSet"]}
				del model_data["metadata"]["vehicle"]["soundSet"]

		if "vehicle" in model_data["metadata"] and vehicletype != "":
			# change placeholder "vehicle" to vehicletype specific value
			model_data["metadata"][vehicletype] = model_data["metadata"]["vehicle"]

		if "vehicle" in model_data["metadata"]:
			del model_data["metadata"]["vehicle"]

		if vehicletype != "":
			# create config table if necessary and make sure it has the right length
			if not "configs" in model_data["metadata"][vehicletype]:
				model_data["metadata"][vehicletype]["configs"] = [{} for i in range(len(model_data["lods"]))]
			elif len(model_data["metadata"][vehicletype]["configs"]) < len(model_data["lods"]):
				for i in range(len(model_data["metadata"][vehicletype]["configs"]), len(model_data["lods"])):
					model_data["metadata"][vehicletype]["configs"].append({})

			configs = model_data["metadata"][vehicletype]["configs"]

			#get potential axles, wheels and co
			if "lods" in model_data:
				for i in range(len(model_data["lods"])):
					if TF_base_functions.export_settings["game"] == "TPF2":
						self.__get_configs_from_group_TPF2(model_data, model_data["lods"][i]["node"]["children"], configs[i])
						if "__configs" in model_data["lods"][i]:
							for key, value in model_data["lods"][i]["__configs"].items():
								configs[i][key] = value
					else:
						self.__get_configs_from_group(model_data, model_data["lods"][i]["children"], configs[i])
						if "__configs" in model_data["lods"][i]:
							for key, value in model_data["lods"][i]["__configs"].items():
								configs[i][key] = value
						
			# for i in range(len(configs)):
				# for key in configs[i]:
					# print("\tconfigs[%i][\"%s\"] = {" % (i, key))
					# for mesh in configs[i][key]:
						# print("\t\t\"%s\"" % mesh)
					# print("\t}")
		else:
			if "transportVehicle" in model_data["metadata"]:
				del model_data["metadata"]["transportVehicle"]
			if "maintenance" in model_data["metadata"]:
				del model_data["metadata"]["maintenance"]
	
	def create_ui_icon(self, data):
		modelData = data["modelData"]
		if "metadata" in modelData and "transportVehicle" in modelData["metadata"]:
			data["icon"] = ("", "")
			if "description" in modelData["metadata"]:
				if "smallIcon" in modelData["metadata"]["description"]:
					path = modelData["metadata"]["description"]["smallIcon"].replace("\\", "/")
					data["smallIcon"] = (path.rpartition("/")[0]+path.rpartition("/")[1], path.rpartition("/")[2].replace(".tga", ""))
				if "icon20" in modelData["metadata"]["description"]:
					path = modelData["metadata"]["description"]["icon20"].replace("\\", "/")
					data["icon20"] = (path.rpartition("/")[0]+path.rpartition("/")[1], path.rpartition("/")[2].replace(".tga", ""))
			
			if "carrier" in modelData["metadata"]["transportVehicle"]:
				if modelData["metadata"]["transportVehicle"]["carrier"] == "RAIL":
					data["iconHeight"] = 56
					data["stageImage"] = "track"
					data["camOffsetZ"] = -0.5
				elif modelData["metadata"]["transportVehicle"]["carrier"] == "TRAM":
					data["iconHeight"] = 48
					data["stageImage"] = "road"
					data["camOffsetZ"] = -0.3
				elif modelData["metadata"]["transportVehicle"]["carrier"] == "ROAD":
					data["iconHeight"] = 40
					data["stageImage"] = "road"
					data["camOffsetZ"] = -0.3
				elif modelData["metadata"]["transportVehicle"]["carrier"] == "AIR":
					data["iconHeight"] = 80
					if "boundingInfo" in modelData and "bbMin" in modelData["boundingInfo"]:
						data["camOffsetZ"] = modelData["boundingInfo"]["bbMin"][2]
				elif modelData["metadata"]["transportVehicle"]["carrier"] == "WATER":
					if "boundingInfo" in modelData and "bbMin" in modelData["boundingInfo"]:
						data["camOffsetZ"] = modelData["boundingInfo"]["bbMin"][2]
					
		else:
			data["icon"] = ("", "")
			data["smallIcon"] = ("", "")
			data["icon20"] = ("", "")

	def load_mesh_always(self, data):
		if TF_base_functions.import_settings["game"] == "TPF2":
			filename2 = TF_base_functions.split_path(data["fileName"])[2]
			fileID = data["modelType"] + filename2
		else:
			fileID = data["modelType"] + data["name"] + ".msh"
		config = ""
		
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport load_mesh_always on %s" % fileID)

		if "metadata_config" in TF_base_functions.import_settings:
			# if TF_base_functions.user_prefs["debug"]:
				# print("Found metadata_config; contents are:")
				# for fileID2 in TF_base_functions.import_settings["metadata_config"]:
					# print("\tkey: %s value: %s" % (fileID2, TF_base_functions.import_settings["metadata_config"][fileID2]))
				# print("Seeking key %s" % fileID)
			if fileID in TF_base_functions.import_settings["metadata_config"]:
				config = TF_base_functions.import_settings["metadata_config"][fileID]
				# if TF_base_functions.user_prefs["debug"]:
					# print("\tMatch on %s" % config)

		configName = config
		addProps = []
		if isinstance(config, tuple):
			configName = config[0]
			addProps = config[1]
		
		if data["meshObj"]:
			if hasattr(data["meshObj"].data.tf_mesh.metadata, configName):
				setattr(data["meshObj"].data.tf_mesh.metadata, configName, True)
			
			for prop, value in addProps:
				if hasattr(data["meshObj"].data.tf_mesh.metadata, prop):
					setattr(data["meshObj"].data.tf_mesh.metadata, prop, value)

	def load_lod(self, data):
		lodData = data["lodData"]
		lod = data["lodObj"]
		model_data = data["modelData"]
		
		if lod and "metadata" in model_data:
			if "vehicle" in model_data["metadata"]:
				# axles, wheels and co
				if "configs" in model_data["metadata"]["vehicle"] and data["lodIndex"] < len(model_data["metadata"]["vehicle"]["configs"]):
					config = model_data["metadata"]["vehicle"]["configs"][data["lodIndex"]]
					for key in _lod_config_keys:
						if key[0] in config:
							key[4](config[key[0]], getattr(lod.tf_lod.metadata, key[0]), lodData["__idToObject"])

	def load_model_data(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport load_model_data")
		
		model_data = data["modelData"]
		#print(model_data)
		errors = []

		if "metadata" in model_data:
			carrier = "NONE"
			if "transportVehicle" in model_data["metadata"]:
				if "carrier" in model_data["metadata"]["transportVehicle"]:
					carrier = model_data["metadata"]["transportVehicle"]["carrier"]
				
				# delete invalid capacities
				if "capacities" in model_data["metadata"]["transportVehicle"]:
					capacities = model_data["metadata"]["transportVehicle"]["capacities"]
					i = 0
					while i < len(capacities):
						capacity = capacities[i]
						if "type" not in capacity or capacity["type"] not in _cargo_types:
							del capacities[i]
						else:
							i += 1
							
			elif "car" in model_data["metadata"]:
				carrier = "ROAD"
				model_data["metadata"]["transportVehicle"] = {"carrier": "ROAD"}

			# get carrier specific information
			vehicletype = ""
			if carrier in _carrier_info:
				vehicletype = _carrier_info[carrier]["vehicletype"]

			if vehicletype != "" and vehicletype in model_data["metadata"]:
				model_data["metadata"]["vehicle"] = model_data["metadata"][vehicletype]
				del model_data["metadata"][vehicletype]
				
				# delete invalid engines
				if vehicletype == "railVehicle":
					if "engines" in model_data["metadata"]["vehicle"]:
						engines = model_data["metadata"]["vehicle"]["engines"]
						i = 0
						while i < len(engines):
							engine = engines[i]
							if "type" not in engine or engine["type"] not in _engine_types:
								del engines[i]
							else:
								i += 1

				# axles, wheels and co
				if "configs" in model_data["metadata"]["vehicle"]:
					if "metadata_config" not in TF_base_functions.import_settings:
						TF_base_functions.import_settings["metadata_config"] = {}

						for i in range(len(model_data["metadata"]["vehicle"]["configs"])):
							config = model_data["metadata"]["vehicle"]["configs"][i]
							for key in _mesh_config_keys:
								if key[0] in config:
									# if (TF_base_functions.user_prefs["debug"]):
										# print("Found key %s in %s" % (key[0], data["fileName"]))
									for mesh in config[key[0]]:
										TF_base_functions.import_settings["metadata_config"][mesh] = key[0]
										# if (TF_base_functions.user_prefs["debug"]):
											# print("Adding key %s to list with value %s" % (mesh, key[0]) )
									del config[key[0]]
			
			# change soundset definition for planes and ships
			if "vehicle" in model_data["metadata"] and carrier in ("WATER", "AIR"):
				if "soundConfig" in model_data["metadata"] and "soundSet" in model_data["metadata"]["soundConfig"]:
					model_data["metadata"]["vehicle"]["soundSet"] = model_data["metadata"]["soundConfig"]["soundSet"]
					del model_data["metadata"]["soundConfig"]
				

	def load_model(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFMetadataImportExport load_model")

		model_data = data["modelData"]

		if "metadata" in model_data:
			if "vehicle" in model_data["metadata"]:
				if "configs" in model_data["metadata"]["vehicle"]:
					for i in range(len(model_data["metadata"]["vehicle"]["configs"])):
						config = model_data["metadata"]["vehicle"]["configs"][i]
						for key in _lod_config_keys:
							if key[0] in config:
								del config[key[0]]
			
			self.__asign_to_prop(model_data["metadata"], data["modelObj"].tf_metadata, "")

class_list = _lod_config_classes + [
	TFMetadataSetting,
	TFAvailabilitySetting,
	TFCostSetting,
	TFDescriptionSetting,
	TFMaintenanceSetting,
	TFParticleSystemSetting,
	TFParticleEmitterSetting,
	TFParticleEmitterAddOp,
	TFParticleEmitterRemoveOp,
	TFParticleEmittersItemUIList,
	TFRailVehicleSetting,
	TFEngineSetting,
	TFEngineAddOp,
	TFEngineRemoveOp,
	TFEngineSettingsItemUIList,
	TFTransportVehicleSetting,
	TFCapacitySetting,
	TFCapacityAddOp,
	TFCapacityRemoveOp,
	TFCapacitySettingsItemUIList,
	TFSoundSetSetting,
	TFCustomMetadataSetting,
	TFCustomMetadataAddOp,
	TFCustomMetadataRemoveOp,
	TFCustomMetadataSettingsItemUIList,
	TFLodMetadataSetting,
	TFMeshMetadataSetting,
]

def register():
	for class_type in class_list:
		bpy.utils.register_class(class_type)

	TF_base_functions.subtype_list_old.append(('engine', 'Engine', '', 3))
	TF_base_functions.subtype_list_old.append(('capacity', 'Capacity', '', 4))
	TF_base_functions.subtype_list_old.append(('particleEmitter', 'Particle Emitter', '', 5))
	TF_base_functions.subtype_list_old.append(('metadata', 'Metadata', '', 128))

	importexport = TFMetadataImportExport()
	TF_base_functions.plugin_add("TFMetadataImportExport", importexport)


def unregister():
	TF_base_functions.plugin_remove("TFMetadataImportExport")
	
	TF_base_functions.subtype_list_old.remove(('engine', 'Engine', '', 3))
	TF_base_functions.subtype_list_old.remove(('capacity', 'Capacity', '', 4))
	TF_base_functions.subtype_list_old.remove(('particleEmitter', 'Particle Emitter', '', 5))
	TF_base_functions.subtype_list_old.remove(('metadata', 'Metadata', '', 128))

	for i in range(len(class_list)-1, -1, -1):
		bpy.utils.unregister_class(class_list[i])