import bpy
import atexit
from mathutils import *
from bpy_extras.io_utils import ExportHelper, ImportHelper
from .tf_util.base_functions import TF_base_functions, basepaths
from .tf_util import model_util, group_util, mesh_util, material_util, image_util, UI_util

# mapping object type to module
modules = {
	"images": image_util,
	"materials": material_util,
	"meshes": mesh_util,
	"groups": group_util,
	"models": model_util,
}
sorted_module_keys = [
	"images",
	"materials",
	"meshes",
	"groups",
	"models",
]
module_keys = [
	"meshes",
	"groups",
	"models",
]

#
error_levels = [
	[
		'DEBUG', 'INFO', 'WARNING', 'ERROR'
	],
	[
		'INFO', 'WARNING', 'ERROR', 'ERROR'
	],
]

user_settings_file = __file__.replace("\\", "/").rpartition("/")[0]+"/user_settings.ini"

###########
# Code starts here, only edit this part when you now what you are doing!
###########

TF_base_functions.read_user_settings(user_settings_file)

def _get_export_types(self, context):
	types_list = []
	for export_type in TF_base_functions.export_file_ext.values():
		types_list.append(export_type)
	return types_list

class TFOperatorProp(bpy.types.PropertyGroup):
	propType = bpy.props.EnumProperty(
		items = [("int", "int", "", 1), ("bool", "bool", "", 2), ("string", "string", "", 3)]
	)
	name = bpy.props.StringProperty()
	propText = bpy.props.StringProperty()
	propDesc = bpy.props.StringProperty()
	propValueInt = bpy.props.IntProperty()
	propValueBool = bpy.props.BoolProperty()
	propValueString = bpy.props.StringProperty()
	
	def setDefault(self, value):
		if self.propType == "int" and "propValueInt" not in self:
			self.propValueInt = value
		elif self.propType == "bool" and "propValueBool" not in self:
			self.propValueBool = value
		elif self.propType == "string" and "propValueString" not in self:
			self.propValueString = value
	
	def draw(self, layout, toggleBool = False):
		if self.propType == "int":
			layout.prop(self, "propValueInt", text = self.propText)
		elif self.propType == "bool":
			layout.prop(self, "propValueBool", text = self.propText, toggle = toggleBool)
		elif self.propType == "string":
			layout.prop(self, "propValueString", text = self.propText)
	
	def getValue(self):
		if self.propType == "int":
			return self.propValueInt
		elif self.propType == "bool":
			return self.propValueBool
		elif self.propType == "string":
			return self.propValueString
	

class TFExport(bpy.types.Operator, ExportHelper):
	"""Train Fever Object Exporter"""
	bl_idname = "object.tf_export"
	bl_label = "Export TF-Object"
	bl_options = {'REGISTER', 'PRESET'}

	check_extension = None
	use_filter_folder = True
	game = bpy.props.EnumProperty(
			name="Game",
			description="Decides which model/mesh format is used.",
			items = [
				('TF', 'Train/Transport Fever', 'Format for use in Train or Transport Fever'),
				('TPF2', 'Transport Fever 2', 'Format for use in Transport Fever 2'),
			],
			default=TF_base_functions.user_settings["export_options"]["game"],
			)
	file_exts = bpy.props.StringProperty(
			default=".msh",
			options={'HIDDEN'},
			)
	filter_glob = bpy.props.StringProperty(
			default="*.mdl;*.grp;*.msh",
			options={'HIDDEN'},
			)
	export_types = bpy.props.EnumProperty(
			items=_get_export_types,
			name="Export filetype",
			description="What filetype should be exported",
			options={'HIDDEN'},
			)
	hidden_faces = bpy.props.BoolProperty(
			name="Hidden faces",
			description="Export hidden faces",
			default=TF_base_functions.user_settings["export_options"]["hidden_faces"]
			)
	unselected_faces = bpy.props.BoolProperty(
			name="Unselected faces",
			description="Export unselected faces",
			default=TF_base_functions.user_settings["export_options"]["unselected_faces"]
			)
	ngons = bpy.props.BoolProperty(
			name="Select Ngons",
			description="Check model for ngons an select them (Exporter only supports Tris and Quads)",
			default=TF_base_functions.user_settings["export_options"]["ngons"]
			)
	flipV = bpy.props.BoolProperty(
			name="Flip V",
			description="Flip V coordinate when exporting",
			default=TF_base_functions.user_settings["export_options"]["flipV"]
			)
	override = bpy.props.BoolProperty(
			name="Override files",
			description="Override existing files",
			default=TF_base_functions.user_settings["export_options"]["override"]
			)
	subfiles = bpy.props.BoolProperty(
			name="Subfiles",
			description="Create all files referenced in this file (.msh, .grp, etc.)",
			default=TF_base_functions.user_settings["export_options"]["subfiles"]
			)
	dir_structure = bpy.props.BoolProperty(
			name="Directory structure",
			description="Store all files in the same structure TF uses (otherwise all files will be stored in the same dir)",
			default=TF_base_functions.user_settings["export_options"]["dir_structure"]
			)
	material_files = bpy.props.BoolProperty(
			name="Material files",
			description="Create material files and textures",
			default=TF_base_functions.user_settings["export_options"]["material_files"]
			)
	ui_icon = bpy.props.BoolProperty(
			name="UI files",
			description="Create both UI icons (only when exporting .mdl)",
			default=TF_base_functions.user_settings["export_options"]["ui_icon"]
			)
	bb = bpy.props.BoolProperty(
			name="Bounding Box",
			description="calculate bounding box from meshes (only when exporting .mdl)",
			default=TF_base_functions.user_settings["export_options"]["bb"]
			)
	mesh_directory = bpy.props.StringProperty(
			name="Mesh dir",
			description="Subdirectory where the .msh files should be stored (only when exporting .mdl)",
			default=TF_base_functions.user_settings["export_options"]["mesh_directory"]
			)
	model_type = bpy.props.EnumProperty(
			items=[(basepaths[i][1], basepaths[i][0], basepaths[i][2], i) for i in range(0, len(basepaths))],
			name="Model type",
			description="Decides where the .mtl file is stored if model_type is empty",
			default=TF_base_functions.user_settings["export_options"]["model_type"]
			)
	prefix = bpy.props.StringProperty(
			name="Prefix",
			description="Prefix prepended to all file names (differentiating them from existing files)",
			default=TF_base_functions.user_settings["export_options"]["prefix"]
			)
	model_type_other = bpy.props.StringProperty(
			name="Custom",
			description="Use this field for other model types (select 'other' in the list above)",
			default=TF_base_functions.user_settings["export_options"]["model_type_other"]
			)
	ui_only = bpy.props.BoolProperty(
			name="UI only",
			description="Only create UI image",
			default=TF_base_functions.user_settings["export_options"]["ui_only"]
			)
	textures = bpy.props.EnumProperty(
			items = [
				("NONE", "No Textures", "Do not create texture files", 0),
				("TGA", "TGA Textures", "Create texture files as .tga", 1),
				#("DDS", "DDS Textures", "Create texture files as .dds (NVidia Plugin required)", 2),
			],
			name ="Textures",
			description = "Sets how the textures files are stored",
			default=TF_base_functions.user_settings["export_options"]["textures"]
			)
	plugin_props = bpy.props.CollectionProperty(type = TFOperatorProp)
	
	def add_prop(self, name, type, text="", description=""):
		if type in ("int", "bool", "string"):
			new_prop = self.plugin_props.add()
			new_prop.propType = type
			new_prop.name = name
			if text != "":
				new_prop.propText = text
			else:
				new_prop.propText = name
			new_prop.propDesc = description
			
			if name in TF_base_functions.user_settings["export_options"]:
				new_prop.setDefault(TF_base_functions.user_settings["export_options"][name])
			
	def _get_plugin_props(self):
		result = {}
		for key in self.plugin_props.keys():
			result[key] = self.plugin_props[key].getValue()
		
		return result

	def __init__(self):
		self.file_ext = ".msh"
		TF_base_functions.call_plugin("init_export_operator", {"operator": self})

	def invoke(self, context, event):
		obj = bpy.context.scene.objects.active
		self.use_filter_folder = True
		
		self.filepath = self.filepath.replace("\\", "/")
		self.filepath = self.filepath.rpartition("/")[0] + self.filepath.rpartition("/")[1]
		
		filename = "Object"
		if obj:
			filename = obj.name

			for type in module_keys:
				if hasattr(modules[type], "is_valid_object") and modules[type].is_valid_object(obj):
					if hasattr(modules[type], "get_export_type"):
						self.export_types = modules[type].get_export_type(obj)
						self.file_exts = "." + self.export_types
					if hasattr(modules[type], "get_model_type"):
						self.model_type, self.model_type_other = modules[type].get_model_type(obj, self.model_type, self.model_type_other)
					if hasattr(modules[type], "get_mesh_dir"):
						self.mesh_directory = modules[type].get_mesh_dir(obj, self.mesh_directory)
					if hasattr(modules[type], "get_filename"):
						filename = modules[type].get_filename(obj)
					if hasattr(modules[type], "create_dir_structure"):
						self.create_dir_structure = modules[type].create_dir_structure(obj)
					break

		self.filepath = self.filepath + filename + self.file_exts
		context.window_manager.fileselect_add(self)
		
		return {'RUNNING_MODAL'}

	def draw(self, context):
		layout = self.layout
		col = layout.column()

		if TF_base_functions.update[0]:
			if TF_base_functions.update[1] != "":
				col.operator("wm.url_open", text="Update available!").url=TF_base_functions.update[1]
			else:
				col.label("Update available!")
			col.separator()

		col.prop(self, "game")
		col.label("Files", icon = 'DOT')
		rowFiles = col.row(True)
		
		if self.export_types != "mod":
			rowFiles.prop(self, "override", text="Override", toggle = True)
			rowFiles.prop(self, "subfiles", text="Subfiles", toggle = True)
			#boxFiles.prop(self, "create_dir_structure")
			#boxFiles.prop(self, "create_material_files")
			col.prop(self, "textures")
			col.separator()

			if self.export_types == "mdl":
				col.label("Icons and Bounding Box", icon = 'DOT')
				rowUI =col.row(True)
				rowUI.prop(self, "ui_icon", text="Icons", toggle = True)
				rowUI.prop(self, "bb", text="BB", toggle = True)
				rowUI.prop(self, "ui_only", text="UI only", toggle = True)
				col.separator()

			col.label("Paths", icon = 'DOT')
			col.prop(self, "mesh_directory")
		else:
			col.separator()
			col.label("Paths", icon = 'DOT')

		col.prop(self, "model_type")
		if self.model_type == "other":
			col.prop(self, "model_type_other")
		col.prop(self, "prefix")
		
		col.separator()
		col.label("Faces", icon = 'DOT')
		rowFaces = col.row(True)
		rowFaces.prop(self, "hidden_faces", text="Hidden", toggle = True)
		rowFaces.prop(self, "unselected_faces", text="Unselected", toggle = True)
		rowFaces.prop(self, "ngons", text="NGons", toggle = True)

		col.separator()
		col.label("UV Mapping", icon = 'DOT')
		col.prop(self, "flipV", text="Flip V Coordinate", toggle = True)
		
		TF_base_functions.call_plugin("draw_export_settings_panel", {"layout": layout, "mydata": self})

	def __create_TF_data_files(self, errors):
		for type in TF_base_functions.export_settings["filedata"].keys():
			if type in modules and hasattr(modules[type], "write_file"):
				for file in TF_base_functions.export_settings["filedata"][type].values():
					if "create" not in file or file["create"]:
						modules[type].write_file(file, errors)

	def export_objects(self, filename, errors):
		model_name = filename

		# get objects
		objects = {}
		for type in module_keys:
			objects[type] = []

		# get all selected objects and group them by type
		for object in bpy.data.objects:
			if object.select:
				object_data = {
					"object": object,
					"name": "",
					"forceOverride": True,
				}
				if object == bpy.context.scene.objects.active:
					object_data["name"] = model_name[:-4]
				if object.parent:
					object_data["parent"] = {"object": object.parent}

				for type in module_keys:
					if hasattr(modules[type], "is_valid_object") and modules[type].is_valid_object(object):
						objects[type].append(object_data)
						break

		basepath_old = TF_base_functions.export_settings["basepath"]
		if model_name == "":
			model_name = bpy.context.scene.objects.active.name
		elif model_name[-3:] in TF_base_functions.export_file_ext.keys():
			n = 6
			if TF_base_functions.export_settings["dir_structure"]:
				i = 0
				found_res = TF_base_functions.export_settings["basepath"].rpartition("/")[2] == "res"
				while not found_res and i < n:
					TF_base_functions.export_settings["basepath"] = TF_base_functions.export_settings["basepath"].rpartition("/")[0]
					found_res = TF_base_functions.export_settings["basepath"].rpartition("/")[2] == "res"
					i = i + 1
				if not found_res:
					TF_base_functions.export_settings["basepath"] = basepath_old + "/res"

		TF_base_functions.export_settings["basepath"] = TF_base_functions.export_settings["basepath"] + "/"
		
		if TF_base_functions.export_settings["ui_icon"]:
			old_scene = bpy.context.screen.scene
			bpy.context.screen.scene = TF_base_functions.export_settings["render_scene"]

		# get data from objects
		for type in module_keys:
			for object in objects[type]:
				if hasattr(modules[type], "create_file_data"):
					modules[type].create_file_data(object, object["name"], errors)

		# create files
		if not TF_base_functions.export_settings["ui_only"]:
			self.__create_TF_data_files(errors)
		
		if TF_base_functions.export_settings["ui_icon"]:
			bpy.context.screen.scene = old_scene

	def execute(self, context):
		# reset addon settings
		TF_base_functions.read_settings()

		# read export settings
		TF_base_functions.export_settings.update(self.as_keywords(ignore=('filter_glob', 'check_existing')))
		TF_base_functions.export_settings.update(self._get_plugin_props())
		
		filepath = TF_base_functions.export_settings["filepath"].replace("\\", "/")
		TF_base_functions.export_settings["basepath"] = filepath.rpartition("/")[0]
		if TF_base_functions.export_settings["model_type"] == 'other':
			TF_base_functions.export_settings["model_type"] = TF_base_functions.export_settings['model_type_other']
		TF_base_functions.export_settings["filedata"] = {}
		TF_base_functions.export_settings["filelist"] = []
		if TF_base_functions.export_settings["ui_only"]:
			TF_base_functions.export_settings["ui_icon"] = True

		#create new scene
		render_scene = bpy.data.scenes.new('UI_small')
		TF_base_functions.export_settings["render_scene_objects"] = UI_util.set_render_scene(render_scene)
		TF_base_functions.export_settings["render_scene"] = render_scene

		filename = filepath.rpartition("/")[2]

		errors = []
		self.export_objects(filename, errors)

		# report errors
		for error in errors:
			level = 0
			if error["level"]:
				level = error["level"]
			if error["message"]:
				self.report({error_levels[TF_base_functions.user_prefs["error_level"]][level]}, error["message"])

		# delete render scene
		UI_util.clear_render_scene(render_scene)
		if "render_scene" in TF_base_functions.export_settings:
			TF_base_functions.export_settings["render_scene"].user_clear()
			bpy.data.scenes.remove(TF_base_functions.export_settings["render_scene"])
			del TF_base_functions.export_settings["render_scene"]

		# clear settings
		TF_base_functions.export_settings.clear()

		return {'FINISHED'}

class TFImport(bpy.types.Operator, ImportHelper):
	"""Train Fever Object Importer"""
	bl_idname = "object.tf_import"
	bl_label = "Import TF-Object"
	bl_options = {'REGISTER', 'UNDO'}

	filename_ext = ".mdl"
	filter_glob = bpy.props.StringProperty(
			default="*.lua;*.mdl;*.grp;*.msh;*.con;*.ani",
			options={'HIDDEN'},
			)
	lod_distance = bpy.props.IntProperty(
			name="Lod distance",
			description="Load lod-model for the specified distance (0 = load all lod-models) (.mdl only)",
			default=TF_base_functions.user_settings["import_options"]["lod_distance"],
			min = 0,
			max = 5000,
			step = 50,
			)
	ui_only = bpy.props.BoolProperty(
			name="UI only",
			description="Only create UI image",
			default=TF_base_functions.user_settings["import_options"]["ui_only"]
			)
	mat_config = bpy.props.IntProperty(
			name="Material Config",
			description="Number of material config to load (.grp and .msh only)",
			min = 0,
			default=TF_base_functions.user_settings["import_options"]["mat_config"]
			)
	flat_faces = bpy.props.BoolProperty(
			name="Flat faces",
			description="Auto detected flat faces (results in less vertices and edges)",
			default=TF_base_functions.user_settings["import_options"]["flat_faces"]
			)
	game = bpy.props.EnumProperty(
			name="Game",
			description="Decides which folder is used to search for required scripts.",
			items = [
				('AUTO', 'auto detected', 'Try to get the game based on the filepath (not recommended)'),
				('TF', 'Train Fever', 'Load required scripts from Train Fever'),
				('TPF', 'Transport Fever', 'Load required scripts from Transport Fever'),
				('TPF2', 'Transport Fever 2', 'Load required scripts from Transport Fever 2'),
			],
			default=TF_base_functions.user_settings["import_options"]["game"],
			)
	material_types = bpy.props.EnumProperty(
			name="Material types",
			description="Material types to create",
			items = [('BLENDER', 'Standard', 'Blender Render without nodes'), ('NODES', 'Nodes', 'Blender Render with nodes'), ('CYCLES', 'Cycles', 'Cycles render with nodes')],
			default={'BLENDER'},
			options={'ENUM_FLAG'},
			)
	at_cursor = bpy.props.BoolProperty(
			name="At Cursor",
			description="Create object at cursor position",
			default=TF_base_functions.user_settings["import_options"]["at_cursor"]
			)
	plugin_props = bpy.props.CollectionProperty(type = TFOperatorProp)
	
	def add_prop(self, name, type, text="", description=""):
		if type in ("int", "bool", "string"):
			new_prop = self.plugin_props.add()
			new_prop.propType = type
			new_prop.name = name
			if text != "":
				new_prop.propText = text
			else:
				new_prop.propText = name
			new_prop.propDesc = description
			
			if name in TF_base_functions.user_settings["import_options"]:
				new_prop.setDefault(TF_base_functions.user_settings["import_options"][name])
			
	def _get_plugin_props(self):
		result = {}
		for key in self.plugin_props.keys():
			result[key] = self.plugin_props[key].getValue()
		
		return result

	def __init__(self):
		TF_base_functions.call_plugin("init_import_operator", {"operator": self})

	def invoke(self, context, event):		
		if context.scene.render.engine == 'CYCLES':
			self.material_types = {'CYCLES'}
		elif context.scene.render.engine == 'BLENDER_GAME':
			self.material_types = {'NODES'}
		else:
			self.material_types = {'BLENDER'}
		
		context.window_manager.fileselect_add(self)
		
		return {'RUNNING_MODAL'}

	def draw(self, context):
		
		layout = self.layout
		col = layout.column()

		if TF_base_functions.update[0]:
			if TF_base_functions.update[1] != "":
				col.operator("wm.url_open", text="Update available!").url=TF_base_functions.update[1]
			else:
				col.label("Update available!")
			col.separator()
		
		rowBools = col.row(True)
		rowBools.prop(self, "ui_only", toggle = True)
		rowBools.prop(self, "flat_faces", toggle = True)
		rowBools.prop(self, "at_cursor", toggle = True)
		col.separator()
		col.prop(self, "lod_distance")
		col.prop(self, "mat_config")
		col.separator()
		col.prop(self, "game")

		TF_base_functions.call_plugin("draw_import_settings_panel", {"layout": layout, "mydata": self})

	def execute(self, context):
		# reset luaLoader and addon settings
		if TF_base_functions.luaLoader:
			del TF_base_functions.luaLoader
			TF_base_functions.luaLoader = None
		TF_base_functions.read_settings()

		# read import settings
		TF_base_functions.import_settings.update(self.as_keywords(ignore=("filter_glob", "check_existing", "plugin_props")))
		TF_base_functions.import_settings.update(self._get_plugin_props())
		
		filepath = TF_base_functions.import_settings["filepath"].replace("\\", "/")
		TF_base_functions.import_settings["basepath"] = ""
		if "/res/" in filepath:
			TF_base_functions.import_settings["basepath"] = filepath.rpartition("/res/")[0]

		if len(TF_base_functions.import_settings["material_types"]) == 0:
			self.report({error_levels[TF_base_functions.user_prefs["error_level"]][2]}, "No material type selected, only placeholder materials will be created.")

		game = TF_base_functions.import_settings["game"]
		if game == "AUTO":
			if "/workshop/content/1066780/" in filepath:
				TF_base_functions.import_settings["game"] = "TPF2"
			elif "/workshop/content/446800/" in filepath:
				TF_base_functions.import_settings["game"] = "TPF"
			elif "/Transport Fever 2/" in filepath:
				TF_base_functions.import_settings["game"] = "TPF2"
			elif "/Transport Fever/" in filepath:
				TF_base_functions.import_settings["game"] = "TPF"
			else:
				TF_base_functions.import_settings["game"] = "TF"
			print(TF_base_functions.import_settings["game"], filepath)

		if TF_base_functions.import_settings["basepath"] == "" and game in TF_base_functions.lua_scripts_basepaths and TF_base_functions.lua_scripts_basepaths[game] != "":
			TF_base_functions.import_settings["basepath"] = TF_base_functions.lua_scripts_basepaths[game]

		filename = filepath.rpartition('/')[2]
		file_ext = filename[-3:]

		errors = []
		
		for object_type in module_keys:
			if hasattr(modules[object_type], "is_valid_import_type") and modules[object_type].is_valid_import_type(file_ext):
				if TF_base_functions.import_settings["ui_only"] and hasattr(modules[object_type], "ui_only_cleanup"):
					for obj in context.scene.objects:
						obj.select = False
					
					modules[object_type].load_from_file(filepath, TF_base_functions.import_settings["mat_config"], True, errors)
					bpy.ops.object.tf_export(
							filepath = filepath,
							ui_only = True,
							bb = (context.scene.objects.active.tf_boundingInfo.bbMin == Vector()) and (context.scene.objects.active.tf_boundingInfo.bbMax == Vector())
						)
					
					modules[object_type].ui_only_cleanup(context)
				
				elif hasattr(modules[object_type], "load_from_file"):
					modules[object_type].load_from_file(filepath, TF_base_functions.import_settings["mat_config"], True, errors)
				break
		
		# report errors
		for error in errors:
			level = 0
			if error["level"]:
				level = error["level"]
			if error["message"]:
				self.report({error_levels[TF_base_functions.user_prefs["error_level"]][level]}, error["message"])

		# clear settings
		if TF_base_functions.luaLoader:
			del TF_base_functions.luaLoader
			TF_base_functions.luaLoader = None
		TF_base_functions.import_settings.clear()

		return {'FINISHED'}		

def get_subtype_enum_callback(self, context):
	return TF_base_functions.subtype_list

def get_subtype_old_enum_callback(self, context):
	return TF_base_functions.subtype_list + TF_base_functions.subtype_list_old
		
# UI Operators
class TFShow_prop(bpy.types.Operator):
	bl_label = "Show property"
	bl_idname = "object.tfprop_show"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		TF_base_functions.user_settings["panels"][self.attr+"_hidden"] = False

		return {'FINISHED'}

class TFHide_prop(bpy.types.Operator):
	bl_label = "Hide property"
	bl_idname = "object.tfprop_hide"

	attr = bpy.props.StringProperty()

	def invoke(self, context, event):
		TF_base_functions.user_settings["panels"][self.attr+"_hidden"] = True

		return {'FINISHED'}

class TFObjectsSettingsPanel(bpy.types.Panel):
	bl_label = "Train Fever Object"
	bl_space_type = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context = "object"

	@classmethod
	def poll(cls, context):
		if (context and context.object and context.object.type in {'ARMATURE', 'EMPTY'}):
			if context.object.subtype in {"model", "mod", "dependency", "group", "lod", "multiple_unit"}:
				return True
		return False

	def draw(self, context):
		layout = self.layout
		if context.object.type in {'EMPTY', 'ARMATURE'}:
			subtype = context.object.subtype
			for object_util in modules.values():
				if hasattr(object_util, "draw_"+subtype):
					getattr(object_util, "draw_"+subtype)(context, layout)

class TFCreate_custom_obj(bpy.types.Operator):
	"""Create special Train Fever object"""
	bl_idname = "object.tf_create_custom_obj"
	bl_label = "Create TF Object"
	bl_options = {'REGISTER', 'UNDO'}

	obj_subtype = bpy.props.EnumProperty(items=get_subtype_enum_callback)

	def execute(self, context):
		options = self.as_keywords()
		subtype = options["obj_subtype"]
		bpy.ops.object.add(type='EMPTY')
		obj = bpy.context.scene.objects.active
		obj.subtype = subtype

		return {'FINISHED'}

def menu_func_export(self, context):
	self.layout.operator(TFExport.bl_idname, text="Train Fever/Transport Fever (.mdl, .grp, .msh)")

def menu_func_import(self, context):
	self.layout.operator(TFImport.bl_idname, text="Train Fever/Transport Fever (.mdl, .grp, .msh)")

def menu_func_custom_props(self, context):
	self.layout.operator_menu_enum(TFCreate_custom_obj.bl_idname, "obj_subtype", text="TF Object", icon='OUTLINER_OB_EMPTY')

class_list = [
	TFOperatorProp,
	TFImport,
	TFExport,
	TFObjectsSettingsPanel,
	TFCreate_custom_obj,
	TFShow_prop,
	TFHide_prop
]

def save_user_settings():
    TF_base_functions.save_user_settings(user_settings_file)
	
def register():
	TF_base_functions.read_settings()
	
	if TF_base_functions.user_prefs["debug"]:
		print("io_mesh_mdl.py register")
	
	for class_type in class_list:
		bpy.utils.register_class(class_type)
		
	bpy.types.INFO_MT_file_import.append(menu_func_import)
	bpy.types.INFO_MT_file_export.append(menu_func_export)
	bpy.types.INFO_MT_add.append(menu_func_custom_props)

	for object_type in sorted_module_keys:
		if hasattr(modules[object_type], "register"):
			modules[object_type].register()

	# general props: subtype and model_type
	bpy.types.Object.subtype = bpy.props.EnumProperty(items=get_subtype_old_enum_callback, options={'HIDDEN'})
	bpy.types.Object.model_type = bpy.props.StringProperty(default="%auto")
	
	atexit.register(save_user_settings)

def unregister():
	if TF_base_functions.user_prefs["debug"]:
		print("tf_mesh_mdl.py unregister")

	# general props: subtype and model_type
	del bpy.types.Object.subtype
	del bpy.types.Object.model_type

	bpy.types.INFO_MT_file_import.remove(menu_func_import)
	bpy.types.INFO_MT_file_export.remove(menu_func_export)
	bpy.types.INFO_MT_add.remove(menu_func_custom_props)

	for i in range(len(class_list)-1, -1, -1):
		bpy.utils.unregister_class(class_list[i])

	for object_type in sorted_module_keys:
		if hasattr(modules[object_type], "unregister"):
			modules[object_type].unregister()
	
	atexit.unregister(save_user_settings)