import bpy
from ..lua import writer
from . import mesh_util
from .base_functions import TF_base_functions
from mathutils import Vector

# Import
def is_valid_import_type(file_ext):
	return file_ext == "grp"

def _validate_data(data, errors):
	TF_base_functions.call_plugin("validate_group_data", data, errors)

	if data["groupData"]:
		# check children
		if "children" in data["groupData"] and str(type(data["groupData"]["children"])) == "<class 'list'>":
			if len(data["groupData"]["children"]) > 0:
				for i in range(0, len(data["groupData"]["children"])):
					child = data["groupData"]["children"][i]
					if str(type(child)) == "<class 'dict'>":
						if "id" not in child or str(type(child["id"])) != "<class 'str'>":
							errors.append({"level": 3, "message": "Missing or wrong field \"children[%i].id\" in Group \"%s\"!" % (i, data["fileName"])})
							return False
						if "type" not in child or str(type(child["type"])) != "<class 'str'>":
							errors.append({"level": 3, "message": "Missing or wrong field \"children[%i].type\" in Group \"%s\"!" % (i, data["fileName"])})
							return False
						if "transf" not in child or str(type(child["transf"])) != "<class 'list'>" or len(child["transf"]) < 16:
							errors.append({"level": 3, "message": "Missing or wrong field \"children[%i].transf\" in Group \"%s\"!" % (i, data["fileName"])})
							return False
					else:
						errors.append({"level": 3, "message": "Field \"children[%i]\" in Group \"%s\" has a wrong data format!" % (i, data["fileName"])})
						return False
				
				# check matConfigs
				if data["matConfig"] >= len(data["groupData"]["matConfigs"]):
					errors.append({"level": 2, "message": "Selected MatConfig index for Group \"%s\" is too high!" % data["fileName"]})
					data["matConfig"] = len(data["groupData"]["matConfigs"])-1
				if len(data["groupData"]["children"]) > len(data["groupData"]["matConfigs"][data["matConfig"]]):
					errors.append({"level": 2, "message": "MatConfig of Group \"%s\" has not enough values!" % data["fileName"]})
					for i in range(len(data["groupData"]["matConfigs"][data["matConfig"]]), len(data["groupData"]["children"])):
						data["groupData"]["matConfigs"][data["matConfig"]].append(0)
				elif len(data["groupData"]["children"]) < len(data["groupData"]["matConfigs"][data["matConfig"]]):
					errors.append({"level": 1, "message": "MatConfig of Group \"%s\" has too much values!" % data["fileName"]})
			else:
				errors.append({"level": 2, "message": "Group \"%s\" has no children!" % data["fileName"]})
				return False
		else:
			errors.append({"level": 3, "message": "Missing or wrong field \"children\" in Group \"%s\"!" % data["fileName"]})
			return False
	else:
		errors.append({"level": 3, "message": "Group \"%s\" could not be loaded or has a wrong data format!" % data["fileName"]})
		return False

	return True

def load_from_file(filename, matConfig, select, errors, select_all = False):
	if (TF_base_functions.user_prefs["debug"]):
		print("load_group_file", filename)

	basepath, filepath, group_name = TF_base_functions.split_path(filename)
	if basepath == "":
		errors.append({"level": 1, "message": "Could not find \"res\"-directory in \"%s\", loading of childrens probably won't be succesfull." %filename})
		basepath = TF_base_functions.import_settings["basepath"] + "/res/"
	
	group_name = group_name.replace(".grp", "")

	# load group data from file
	group_data = TF_base_functions.get_data(basepath + filepath, "GROUP", errors)
	data = {
		"groupData": group_data,
		"fileName": filename,
		"matConfig": matConfig,
		"name": group_name,
		"basepath": basepath,
	}
	TF_base_functions.call_plugin("load_group_data", data, errors)

	if _validate_data(data, errors):
		# create group object
		new_group = bpy.data.objects.new(group_name, None)
		new_group.subtype = "group"
		new_group.model_type = TF_base_functions.get_model_type(filename, "models/group/")
		new_group.select = select
		new_parent = new_group
		scene = bpy.context.scene
		scene.objects.link(new_group)

		# load children
		i = 0
		id_to_object = []
		for child in group_data["children"]:
			child_filename = child["id"]
			child_obj = None
			if child["type"] == "MESH":
				child_obj = mesh_util.load_from_file(basepath + "models/mesh/" + child_filename, group_data["matConfigs"][matConfig][i], select_all, errors)
				id_to_object += [(child_obj, child_filename)]
			elif child["type"] == "GROUP":
				child_obj, ids = load_from_file(basepath + "models/group/" + child_filename, group_data["matConfigs"][matConfig][i], select_all, errors, select_all)
				id_to_object += [(child_obj, child_filename)] + ids
			else:
				id_to_object += [(child_obj, child_filename)]
			
			if child_obj:
				child_obj.matrix_local = TF_base_functions.create_4x4matrix(child["transf"])
				child["blender_object"] = child_obj
				if new_parent:
					child_obj.parent = new_parent
			i = i + 1

		data["groupObj"] = new_group
		if TF_base_functions.import_settings["at_cursor"]:
			new_group.location = bpy.context.scene.cursor_location
		TF_base_functions.call_plugin("load_group", data, errors)

		return new_group, id_to_object
	
	return None, []

# Export
def is_valid_object(object):
	return object.type in {'EMPTY', 'ARMATURE'} and object.subtype == "group"

def get_export_type(object):
	return "grp"

def get_model_type(object, model_type, model_type_other):
	return TF_base_functions.get_modeltype(object.model_type, model_type, model_type_other)

def create_file_data_TPF2(group_object, name, errors):
	group_obj = group_object["object"]

	# get name
	group_name = name
	if group_name == "":
		group_name = group_obj.name.partition(".")[0]

	if (TF_base_functions.user_prefs["debug"]):
		print("save_group_file", group_name)

	if "groups" not in TF_base_functions.export_settings["filedata"]:
		TF_base_functions.export_settings["filedata"]["groups"] = {}
	filedata = TF_base_functions.export_settings["filedata"]["groups"]

	# get model type
	model_type = group_obj.model_type
	if model_type == "%auto":
		model_type = TF_base_functions.export_settings["model_type"]
	elif model_type == "%inherit" and "parent" in group_object and "model_type" in group_object["parent"]:
		model_type = group_object["parent"]["model_type"]

	file_id = model_type + group_name
	create_group = TF_base_functions.export_settings["subfiles"] or ("forceOverride" in group_object and group_object["forceOverride"])
	bounding_box = None
	if TF_base_functions.export_settings["bb"]:
		bounding_box = {"bbMax": [0.0, 0.0, 0.0], "bbMin": [0.0, 0.0, 0.0]}

	if file_id in filedata:
		group_data = filedata[file_id]["data"]
	else:
		group_data = {"children": [], "matConfigs": []}

	# check children
	matConfig = []
	possible_indices = set(range(len(group_data["matConfigs"])))
	for child in group_obj.children:
		matConfig_id = -1
		bound_box = None
		transf = []
		for i in range(4):
			for j in range(4):
				transf.append(round(child.matrix_local.col[i][j], 5))

		child_name = ""
		child_type = ""
		child_id = ""

		if child.type == 'MESH':
			child_type = "MESH"
			child_name = child.data.name

			# get model_type and mesh dir for child
			child_model_type = child.data.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type

			child_mesh_dir = child.data.mesh_dir
			if child_mesh_dir == "%auto":
				child_mesh_dir = TF_base_functions.export_settings["mesh_directory"]
			child_id = child_model_type + child_mesh_dir + "/"[:len(child_mesh_dir)] + child_name + ".msh"

			# get matConfig an create child data
			mesh_object = {
				"object": child,
				"parent": {"object": group_obj, "model_type": model_type}
			}
			matConfigID, bound_box = mesh_util.create_file_data(mesh_object, "", errors)

		elif child.type in {'EMPTY', 'ARMATURE'}:
			child_type = "GROUP"
			child_name = child.name.partition(".")[0]

			# get model_type for child
			child_model_type = child.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type
			child_id = child_model_type + child_name + ".grp"

			# get matConfig an create child data
			grp_object = {
				"object": child,
				"parent": {"object": group_obj, "model_type": model_type}
			}
			matConfigID, bound_box = create_file_data(grp_object, "", errors)

		else:
			# invalid child
			errors.append({"level": 1, "message": "Invalid type for child \"%s\" of group \"%s\", this child will be ignored!" %(child.name, group_obj.name)})

		# BB calculation
		if bounding_box and bound_box:
			for key in {"bbMin", "bbMax"}:
				for i in range(3):
					vec = child.matrix_local*Vector(bound_box[key])
					if vec[i] < bounding_box["bbMin"][i]:
						bounding_box["bbMin"][i] = vec[i]
					if vec[i] > bounding_box["bbMax"][i]:
						bounding_box["bbMax"][i] = vec[i]


		if child_id and child_type and {"id": child_id, "transf": transf, "type": child_type} not in group_data["children"]:
			group_data["children"].append({"id": child_id, "transf": transf, "type": child_type})

	data = {
		"groupData": group_data,
		"groupObj": group_obj,
		"name": group_name,
		"fileID": file_id,
	}
	TF_base_functions.call_plugin("save_group_data", data, errors)

	filedata[file_id] = {"data": group_data, "model_type": model_type, "name": group_name, "create": create_group}
	if "forceOverride" in group_object:
		filedata[file_id]["forceOverride"] = group_object["forceOverride"]
	return bounding_box

def create_file_data(group_object, name, errors):
	group_obj = group_object["object"]

	# get name
	group_name = name
	if group_name == "":
		group_name = group_obj.name.partition(".")[0]

	if (TF_base_functions.user_prefs["debug"]):
		print("save_group_file", group_name)

	if "groups" not in TF_base_functions.export_settings["filedata"]:
		TF_base_functions.export_settings["filedata"]["groups"] = {}
	filedata = TF_base_functions.export_settings["filedata"]["groups"]

	# get model type
	model_type = group_obj.model_type
	if model_type == "%auto":
		model_type = TF_base_functions.export_settings["model_type"]
	elif model_type == "%inherit" and "parent" in group_object and "model_type" in group_object["parent"]:
		model_type = group_object["parent"]["model_type"]

	file_id = model_type + group_name
	create_group = TF_base_functions.export_settings["subfiles"] or ("forceOverride" in group_object and group_object["forceOverride"])
	bounding_box = None
	if TF_base_functions.export_settings["bb"]:
		bounding_box = {"bbMax": [0.0, 0.0, 0.0], "bbMin": [0.0, 0.0, 0.0]}

	if file_id in filedata:
		group_data = filedata[file_id]["data"]
	else:
		group_data = {"children": [], "matConfigs": []}

	# check children
	matConfig = []
	possible_indices = set(range(len(group_data["matConfigs"])))
	for child in group_obj.children:
		matConfig_id = -1
		bound_box = None
		transf = []
		for i in range(4):
			for j in range(4):
				transf.append(round(child.matrix_local.col[i][j], 5))

		child_name = ""
		child_type = ""
		child_id = ""

		if child.type == 'MESH':
			child_type = "MESH"
			child_name = child.data.name

			# get model_type and mesh dir for child
			child_model_type = child.data.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type

			child_mesh_dir = child.data.mesh_dir
			if child_mesh_dir == "%auto":
				child_mesh_dir = TF_base_functions.export_settings["mesh_directory"]
			child_id = child_model_type + child_mesh_dir + "/"[:len(child_mesh_dir)] + child_name + ".msh"

			# get matConfig an create child data
			mesh_object = {
				"object": child,
				"parent": {"object": group_obj, "model_type": model_type}
			}
			matConfig_id, bound_box = mesh_util.create_file_data(mesh_object, "", errors)

		elif child.type in {'EMPTY', 'ARMATURE'}:
			child_type = "GROUP"
			child_name = child.name.partition(".")[0]

			# get model_type for child
			child_model_type = child.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type
			child_id = child_model_type + child_name + ".grp"

			# get matConfig an create child data
			grp_object = {
				"object": child,
				"parent": {"object": group_obj, "model_type": model_type}
			}
			matConfig_id, bound_box = create_file_data(grp_object, "", errors)

		else:
			# invalid child
			errors.append({"level": 1, "message": "Invalid type for child \"%s\" of group \"%s\", this child will be ignored!" %(child.name, group_obj.name)})

		if matConfig_id == -1:
			matConfig_id = 0
		matConfig.append(matConfig_id)

		# BB calculation
		if bounding_box and bound_box:
			for key in {"bbMin", "bbMax"}:
				for i in range(3):
					vec = child.matrix_local*Vector(bound_box[key])
					if vec[i] < bounding_box["bbMin"][i]:
						bounding_box["bbMin"][i] = vec[i]
					if vec[i] > bounding_box["bbMax"][i]:
						bounding_box["bbMax"][i] = vec[i]


		if child_id and child_type and {"id": child_id, "transf": transf, "type": child_type} not in group_data["children"]:
			group_data["children"].append({"id": child_id, "transf": transf, "type": child_type})

	matConfigID = len(group_data["matConfigs"])
	if matConfig in group_data["matConfigs"]:
		matConfigID = group_data["matConfigs"].index(matConfig)
	else:
		group_data["matConfigs"].append(matConfig)

	data = {
		"groupData": group_data,
		"groupObj": group_obj,
		"matConfigID": matConfigID,
		"name": group_name,
		"fileID": file_id,
	}
	TF_base_functions.call_plugin("save_group_data", data, errors)

	filedata[file_id] = {"data": group_data, "model_type": model_type, "name": group_name, "create": create_group}
	if "forceOverride" in group_object:
		filedata[file_id]["forceOverride"] = group_object["forceOverride"]
	return matConfigID, bounding_box

def write_file(file, errors):
	filename = file["name"] + ".grp"
	filepath = file["model_type"]

	id = filepath + filename

	resfilepath = "models/group/" + filepath
	resfilename = resfilepath + filename

	if TF_base_functions.export_settings["dir_structure"]:
		osfilepath = TF_base_functions.export_settings["basepath"]+resfilepath
	else:
		osfilepath = TF_base_functions.export_settings["basepath"]

	osfilename = osfilepath + filename

	if (not writer.isfile(osfilename)) or TF_base_functions.export_settings["override"] or ("forceOverride" in file and file["forceOverride"]):
		if not writer.exists(osfilepath):
			writer.makedirs(osfilepath)

		filetext = ""
		# additional code, to do: not restricted to require
		if "__requires" in file["data"]:
			for req in file["data"]["__requires"]:
				filetext += req + '\n'

		filetext += "function data()\nreturn "
		filetext += writer.write_var_struct_string(file["data"], "", 0, True, False)
		filetext += "end"

		data = {
			"text": filetext,
			"groupData": file["data"],
			"fileID": id,
		}
		TF_base_functions.call_plugin("write_group_file", data, errors)

		output = open(osfilename, 'w', encoding="utf-8")
		output.write(data["text"])
		output.close()

	TF_base_functions.export_settings["filelist"].append([osfilename, "res/" + resfilename])

# UI functions
def draw_group(context, layout):
	mydata = context.object

	if mydata:
		TF_base_functions.call_plugin("draw_group_settings_panel", {"layout": layout, "mydata": mydata})
		TF_base_functions.draw_internal_box(mydata, layout, "group")
		

def register():
	TF_base_functions.subtype_list.append(('group', 'Group', '', 1))
	
	TF_base_functions.export_file_ext["grp"] = (("grp", "Group (.grp)", "", 2))

def unregister():
	TF_base_functions.subtype_list.remove(('group', 'Group', '', 1))
	
	del TF_base_functions.export_file_ext["grp"]